/* startup stuff */

/* it should be possible to throw away the text/data/bss of the object
   file resulting from this source -- so, we don't define here
   anything we could still use at a later time.  instead, globals are
   defined in globals.c */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>

#include <l4/sys/syscalls.h>
#include <l4/sys/ipc.h>
#include <l4/sys/kdebug.h>

#include "config.h"
#include "memmap.h"
#include "globals.h"
#include "exec.h"
#include "rmgr.h"
#include "cfg.h"
#include "symbols.h"
#include "lines.h"

#include <l4/rmgr/proto.h>
#include "init.h"

int l4_version = 0;			/* L4 version number (startup.c) */
int boot_errors = 0;			/* no. of boot errors (startup.c) */
int boot_wait = 0;			/* wait after boot? (cfg-parse.y) */
int no_symbols = 0;			/* 1=don't load module symbols */
int no_lines = 0;			/* 1=don't load module stab info */
bootquota_t bootquota[TASK_MAX];	/* boot quotas (cfg-parse.y) */
l4util_mb_vbe_ctrl_t mb_vbe_ctrl;	/* VESA contr. info (startup.c) */
l4util_mb_vbe_mode_t mb_vbe_mode;	/* VESA mode info (startup.c) */
l4util_mb_mod_t mb_mod[MODS_MAX];	/* multiboot modules (startup.c) */

/* Private declarations */
static void init_globals(void);
static void init_memmap(void);
static void init_iomap(void);
static void init_rmgr_debug_info(void);
#ifdef FIASCO_UX
static void init_noloader(void);
#endif
static void init_irq(void);
static void init_task(void);
static void init_small(void);
static void configure(void);
static void start_tasks(void);

static void setup_task(l4_threadid_t t);
static int  overlaps_myself(vm_offset_t begin, vm_offset_t end);
static void reserve_module_memory(vm_offset_t start,
                                  vm_offset_t end, int taskno);
static int  alloc_exec_read_exec(void *handle,
				 vm_offset_t file_ofs, vm_size_t file_size,
				 vm_offset_t mem_addr, vm_size_t mem_size,
				 exec_sectype_t section_type);
static int  alloc_exec_read(void *handle, vm_offset_t file_ofs,
			    void *buf, vm_size_t size,
			    vm_size_t *out_actual);
static int  set_small(l4_threadid_t t, int num);
static int  set_prio(l4_threadid_t t, int num);

static int  error_stop = 1;	/* wait for keypress on boot error? */

/* return module name */
static const char*
mb_mod_name(int i)
{
  return (i<0 || i>=MODS_MAX || !mb_mod[i].cmdline)
    ? "[unknown]"
    : (char *)mb_mod[i].cmdline;
}

/* a boot error occured (shared with cfg-parse.y) */
void
boot_error(void)
{
  if (error_stop)
    {
      printf("\r\nRMGR: boot error. "
	     "Return continues, Esc panics, \"m\" shows memory map");
#ifdef FIASCO_UX
      puts("");
      show_mem_info();
      enter_kdebug("RMGR: boot error");
#else
	{
	  char c = getchar();
	  printf("\r%60s\r","");
	  switch (c)
	    {
	    case 0x1B:
	      panic("RMGR: boot error");
	    case 'm':
	    case 'M':
	      show_mem_info();
	      break;
	    }
	}
#endif
    }
  boot_errors++;
}

/* a fatal boot error occured, don't allow to continue */
static void
boot_fatal(void)
{
  for (;;)
    {
      printf("\r\nRMGR: Fatal boot error. "
	     "Esc panics, \"m\" shows memory map");
#ifdef FIASCO_UX
      puts("");
      show_mem_info();
      enter_kdebug("RMGR: fatal boot error");
#else
      for (;;)
	{
	  char c = getchar();
	  switch (c)
	    {
	    case 0x1B:
	      printf("\r%70s\r","");
	      panic("RMGR: fatal boot error");
	    case 'm':
	    case 'M':
	      printf("\r%70s\r","");
	      show_mem_info();
	      break;
	    default:
	      continue;
	    }
	  break;
	}
#endif
    }
}

static void
boot_panic(const char *fmt, ...)
{
  for (;;)
    {
      va_list args;
      va_start(args, fmt);
      vprintf(fmt, args);
      va_end(args);
      printf("\r\nRMGR: Return panics, \"m\" shows memory map");
#ifdef FIASCO_UX
      puts("");
      show_mem_info();
      enter_kdebug("RMGR: panic");
#else
	{
	  char c = getchar();
	  printf("\r%60s\r","");
	  switch (c)
	    {
	    case 0x0D:
	      panic("RMGR: panic");
	    case 'm':
	    case 'M':
	      show_mem_info();
	      break;
	    }
	}
#endif
    }
}

/* started as the L4 booter task through startup() in startup.c */
void
init(void)
{
  vm_offset_t address;

  printf("RMGR: Hi there!\n");

#ifdef FIASCO_UX
  init_noloader();
#endif

  init_globals();

  /* Make sure that loading symbols of rmgr, fiasco or other modules doesn't
   * overwrite the raw modules in memory. Don't care about module 0 (rmgr)
   * and module 1 (sigma0) */
  mod_exec_init(mb_mod, 2, mb_info.mods_count);

  printf("RMGR: bootloader loaded %u module%s at %08x-%08x\n",
          mb_info.mods_count - first_task_module,
          "s" + (mb_info.mods_count - first_task_module == 1),
          mod_range_start, mod_range_end);

  init_memmap();

  init_rmgr_debug_info();
  init_iomap();
  init_irq();
  init_task();
  init_small();

  configure();

  setup_task(myself);
  setup_task(my_pager);

  /* start the tasks loaded as modules */
  start_tasks();

  /* add the memory used by this module and not any longer required
     (the ".init" section) to the free memory pool */
  for (address = l4_trunc_page(&_start);
       address < l4_trunc_page(&_stext);
       address += L4_PAGESIZE)
    {
      check(memmap_free_page(address, O_RESERVED));
    }

  if (boot_errors || boot_wait)
    {
      if (boot_errors)
	printf("RMGR: WARNING: there were %d boot errors -- continue?\n"
	       "               Return continues, Esc panics, "
	       "\"k\" enters L4 kernel debugger...\n",
	       boot_errors);
      else
	printf("RMGR: bootwait option was set -- please press a key\n"
	       "      Return continues, Esc panics, "
	       "\"k\" enters L4 kernel debugger...\n");
#ifdef FIASCO_UX
      enter_kdebug("boot error, you can go on");
#else
	{
	  char c = getchar();

	  if (c == 27)
	    panic("RMGR: boot error");
	  else if (c == 'k' || c == 'K')
	    enter_kdebug("boot error");
	}
#endif
    }


  /* now start the resource manager */
  rmgr();
}

/* support functions for init() */

static
void configure(void)
{
  char *config, *config_end;

  cfg_init();

  /* find the config data */
  if (! (mb_info.flags & L4UTIL_MB_CMDLINE))
    return;

  config = strchr((char *) mb_info.cmdline, ' ');
  if (! config)
    return;

  config++;			/* skip ' ' character */
  if (!strstr(config, "-configfile"))
    {
      /* not using a configfile -- use the cmdline */
      config_end = config + strlen(config);

      /* skip options */
      while (config < config_end)
	{
	  if (*config == ' ')	/* skip spaces */
	    {
	      config++;
	      continue;
	    }

	  if (*config != '-')	/* end of options found */
	    break;

	  config = strchr(config, ' ');	/* find end of this option */
	  if (!config)
	    return;
	}
    }
  else
    {
      /* the configfile is in the first module... */

      /* make sure the config file is not the last module */
      check(mb_info.mods_count >= first_task_module + 2);

      config = (char *) mb_mod[first_task_module].mod_start;
      config_end = (char *) mb_mod[first_task_module].mod_end;

      /* is it really a config file? */
      if (strncmp(config, "#!rmgr", 6) != 0)
	{
	  printf("RMGR: ERROR: 2nd module is not a config file "
		 "(doesn't start with \"#!rmgr\")\n");
	  boot_error();
	  return;
	}

      first_task_module++;
    }

  cfg_setup_input(config, config_end);

  if (cfg_parse() != 0)
    boot_error();

  printf("RMGR: log_mask: %08x, log_types: %08x\n", 
         debug_log_mask, debug_log_types);

  /* a few post-parsing actions follow... */
  if (small_space_size)
    {
      vm_offset_t s = small_space_size;

      small_space_size /= 0x400000; /* divide by 4MB */
      if (s > small_space_size * 0x400000 /* did we delete any size bits? */
	  || (small_space_size & (small_space_size - 1))) /* > 1 bit set? */
	{
	  /* use next largest legal size */
	  int count;
	  for (count = 0; small_space_size; count++)
	    {
	      small_space_size >>= 1;
	    }

	  small_space_size = 1L << count;
	}
      
      if (small_space_size > 64)
	{
	  printf("RMGR: ERROR: small_space_size 0x%08x too large\n", s);
	  small_space_size = 0;
	  boot_error();
	}

      else
	{
	  unsigned i;

	  for (i = 1; i < 128/small_space_size; i++)
	    check(small_free(i, O_RESERVED));

	  printf("RMGR: configured small_space_size = %d MB\n",
		 small_space_size * 4);
	}
    }
}

static
void setup_task(l4_threadid_t t)
{
  unsigned me = t.id.task;
  bootquota_t *b = & bootquota[me];

  if (b->small_space != 0xff)
    {
      if (small_space_size)
	set_small(t, b->small_space);
      else
	b->small_space = 0xFF;
    }

  set_prio(t, b->prio);
}

#ifdef FIASCO_UX
static
void init_noloader(void)
{
  int i;
  char *modname;

  l4_version = VERSION_FIASCO;

  /* Set up multiboot info structure inside rmgr from mbi ptr */
  mb_info = *mb_ptr;

  for (i = 0; i < mb_info.mods_count; i++)
    {
      /* Set up multiboot modules array */
      memcpy (mb_mod + i, ((l4util_mb_mod_t *) mb_info.mods_addr) + i,
              sizeof (*mb_mod));
    
      modname = (char *) mb_mod[i].cmdline;

      /* Setup mod_names string array */
      snprintf (mb_mod_names[i], sizeof (*mb_mod_names), "%s", modname);
    
      /* Repoint module command line string */
      mb_mod[i].cmdline = (unsigned) mb_mod_names[i];
    }

  /* Setup RMGR cmdline */
  snprintf (mb_cmdline, sizeof (mb_cmdline), "%s", (char *) mb_mod[1].cmdline);
  mb_info.cmdline = (vm_offset_t) mb_cmdline;

  if (mb_info.flags & L4UTIL_MB_CMDLINE)
    {
      if (strstr ((char *) mb_info.cmdline, " -symbols"))
        {
          fiasco_symbols     = mb_mod[first_task_module].mod_start;
          fiasco_symbols_end = l4_round_page(mb_mod[first_task_module].mod_end);
          first_task_module++;
        }

      if (strstr ((char *) mb_info.cmdline, " -lines"))
        {
          fiasco_lines       = mb_mod[first_task_module].mod_start;
          fiasco_lines_end   = l4_round_page(mb_mod[first_task_module].mod_end);
          first_task_module++;
        }
    }

  if (mb_info.flags & L4UTIL_MB_VIDEO_INFO)
    {
      l4util_mb_info_t *mbi = (l4util_mb_info_t *) &mb_info;

      /* copy VBE mode info structure */
      if (mbi->vbe_mode_info)
	{
	  memcpy(&mb_vbe_mode, (void*)(mbi->vbe_mode_info),
	         sizeof(l4util_mb_vbe_mode_t));
	  mb_info.vbe_mode_info = (vm_offset_t) &mb_vbe_mode;
	}
      
      /* copy VBE controller info structure */
      if (mbi->vbe_ctrl_info)
	{
	  memcpy(&mb_vbe_ctrl, (void*)(mbi->vbe_ctrl_info),
	         sizeof(l4util_mb_vbe_ctrl_t));
	  mb_info.vbe_ctrl_info = (vm_offset_t) &mb_vbe_ctrl;
	}
    }
}
#endif

static
void init_globals(void)
{
  l4_umword_t dummy;

  /* set some globals */

  /* set myself (my thread id) */
  myself = rmgr_pager_id = rmgr_super_id = l4_myself();
  rmgr_pager_id.id.lthread = RMGR_LTHREAD_PAGER;
  rmgr_super_id.id.lthread = RMGR_LTHREAD_SUPER;

  /* set my_pager */
  my_preempter = my_pager = L4_INVALID_ID;
  l4_thread_ex_regs(myself, (l4_umword_t) -1, (l4_umword_t) -1, 
		    &my_preempter, &my_pager, &dummy, &dummy, &dummy);

  /* set mem size */
  mem_upper = mb_info.mem_upper;
  mem_lower = mb_info.mem_lower & ~3; /* round down to next 4k boundary */

  /* more global initialization stuff */
  small_space_size = 0;

  debug = verbose = debug_log_mask = debug_log_types = 0;

#if 0
  if (no_pentium == 1)
    printf("RMGR: running on L4/486\n");
  else if (no_pentium == 0)
    printf("RMGR: running on L4/Pentium\n");
  else {
    printf("RMGR: unknown L4 version, assuming L4/486\n");
    no_pentium = 1;
  }
#else
  printf("RMGR: running on %s\n", KERNEL_NAMES[l4_version]);
#endif

  if (mb_info.flags & L4UTIL_MB_CMDLINE)
    {
      if (strstr((char *) mb_info.cmdline, "-nopentium"))
	{
	  no_pentium=1;
	  printf("      no l4/pentium\n");
	}
      if (strstr((char *) mb_info.cmdline, "-nocheckdpf"))
	{
	  no_checkdpf=1;
	  printf("      don't check double page faults\n");
	}
      if (strstr((char *) mb_info.cmdline, "-errorstop"))
	error_stop=1;
      if (strstr((char *) mb_info.cmdline, "-memdump"))
	boot_mem_dump=1;
      if (strstr((char *) mb_info.cmdline, "-nosymbols"))
	{
	  no_symbols=1;
	  printf("      don't load any symbol information from modules\n");
	}
      if (strstr((char *) mb_info.cmdline, "-nolines"))
	{
	  no_lines=1;
	  printf("      don't load any lines information from modules\n");
	}
    }
  quota_init();
}

static
void init_irq(void)
{
  int err, i;
  unsigned *sp;
  l4_umword_t code, dummy;
  l4_msgdope_t result;
  l4_threadid_t t, preempter, pager;

  /* initialize to "reserved" */
  irq_init();

  printf("RMGR: attached irqs = [ ");

  /* start a thread for each irq */
  t = myself;
  
  for (i = 0; i < IRQ_MAX; i++)
    {
      t.id.lthread = LTHREAD_NO_IRQ(i);
      sp = (unsigned *)(__irq_stacks + (i + 1) * __IRQ_STACKSIZE);

      *--sp = i;		/* pass irq number as argument to thr func */
      *--sp = 0;		/* faked return address */
      preempter = my_preempter; pager = my_pager;
      l4_thread_ex_regs(t, (l4_umword_t) __irq_thread, (l4_umword_t) sp,
                        &preempter, &pager, &dummy, &dummy, &dummy);

      /* handshake and status check */
      err = l4_ipc_receive(t, L4_IPC_SHORT_MSG, &code, &dummy,
                           L4_IPC_NEVER, &result);

      assert(err == 0);
      
#ifndef FIASCO_UX
      if (i == 2)
        {
          /* enable IRQ2 */
          __asm__("inb $0x21,%al\n\t"
                  "jmp 1f\n\t"
                  "1:\tjmp 1f\n\t"
                  "1:\tandb $0xfb,%al\n\t"
                  "outb %al,$0x21\n\t");
        }

      if (code == 0)
        {
          if (i != 2) 
	    irq_free(i, O_RESERVED); /* free irq, except if irq2 */
          printf("%x ", i);
        }
#else      
      if (code == 0) {
	irq_free(i, O_RESERVED);
	printf("%x ", i);
      }
#endif
      else
	printf("<!%x> ", i);
    }
  printf("]\n");
}

static void
init_task(void)
{
  unsigned i;

  task_init();

  for (i = myself.id.task + 1; i < TASK_MAX; i++)
    __task[i] = O_FREE;

  first_not_bmod_free_modnr = mb_info.mods_count;
  first_not_bmod_free_task  = first_not_bmod_task
			    = myself.id.task + 1
			    + (mb_info.mods_count-first_task_module);
}

static void
init_small(void)
{
  small_init();
}


static void
init_memmap(void)
{
  l4_msgdope_t result;
  l4_snd_fpage_t sfpage;
  vm_offset_t address, a;
  vm_size_t free_size = 0, reserved_size = 0;
  int error;
  int ignore_high_pages = 0;
  
  /* initialize memory to "reserved" */
  memmap_init();

  /* invalidate the ``last page fault'' array */
  for (a = 0; a < TASK_MAX; a++)
    {
      last_pf[a].pfa = -1;
      last_pf[a].eip = -1;
    }

  mem_high = 0;

  if (!no_pentium) 
    {
      /* get as much super (4MB) pages as possible, starting with
     superpage 1 (0x400000) */
      for (address = L4_SUPERPAGESIZE; address < MEM_MAX; 
	   address += L4_SUPERPAGESIZE)
	{
	  if (address & (L4_SUPERPAGESIZE-1))
	    {
	      error = l4_ipc_call(my_pager, L4_IPC_SHORT_MSG,
				  address | 1, 0,
				  L4_IPC_MAPMSG(0, L4_WHOLE_ADDRESS_SPACE),
				  &sfpage.snd_base, &sfpage.fpage.fpage,
				  L4_IPC_NEVER, &result);
	    }
	  else
	    {
	      error = l4_ipc_call(my_pager, L4_IPC_SHORT_MSG,
				  address | 1, L4_LOG2_SUPERPAGESIZE << 2,
				  L4_IPC_MAPMSG(0, L4_WHOLE_ADDRESS_SPACE),
				  &sfpage.snd_base, &sfpage.fpage.fpage,
				  L4_IPC_NEVER, &result);
	    }
	  /* XXX should i check for errors? */

	  if (! (l4_ipc_fpage_received(result) 
		 && l4_ipc_is_fpage_writable(sfpage.fpage)
		 && sfpage.fpage.fp.size == L4_LOG2_SUPERPAGESIZE))
	    {
	      vm_size_t size;

	      if (!l4_ipc_fpage_received(result))
		continue;
	      
	      size = 1L << sfpage.fpage.fp.size;
	      assert(size == l4_trunc_page(size));
	      
	      for (; size > 0; size -= L4_PAGESIZE, address += L4_PAGESIZE)
		{
		  free_size += L4_PAGESIZE;
		  memmap_free_page(address, O_RESERVED);
		}
	     
	      if (mem_high < address)
    		mem_high = address;

	      /* compensate next loop iteration */
	      address -= L4_SUPERPAGESIZE;
	      continue;
	    }

	  /* kewl, we've received an super awsum page */
	  free_size += L4_SUPERPAGESIZE;

	  for (a = address; a < address + L4_SUPERPAGESIZE; a += L4_PAGESIZE)
	    memmap_free_page(a, O_RESERVED);

	  if (mem_high < a)
	    mem_high = a;

	  assert(memmap_owner_superpage(address) == O_FREE);
	}
    }
  /* page in BIOS data area page explicitly.  like adapter space it will be
     marked "reserved" and special-cased in the pager() in memmap.c */
  error = l4_ipc_call(my_pager, L4_IPC_SHORT_MSG,
		      0, 0,
		      L4_IPC_MAPMSG(0, L4_WHOLE_ADDRESS_SPACE),
		      &sfpage.snd_base, &sfpage.fpage.fpage,
		      L4_IPC_NEVER, &result);
  /* XXX should i check for errors? */

  /* now add all memory we can get from L4 to the memory map as free
     memory */
  for (;;)
    {
      error = l4_ipc_call(my_pager, L4_IPC_SHORT_MSG,
			  0xfffffffc, 0,
			  L4_IPC_MAPMSG(0, L4_WHOLE_ADDRESS_SPACE),
			  &sfpage.snd_base, &sfpage.fpage.fpage,
			  L4_IPC_NEVER, &result);
      /* XXX should i check for errors? */

      if (! sfpage.fpage.fpage)
	break;			/* out of memory */

      if (sfpage.snd_base >= MEM_MAX)
	{
	  if (!ignore_high_pages)
	    {
	      ignore_high_pages = 1;
	      printf("RMGR: WARNING: all memory above %ld MB is wasted!\n",
		     MEM_MAX/(1024*1024));
	      boot_error();
	    }
	  continue;
	}

      if (mem_high < sfpage.snd_base + L4_PAGESIZE) 
	mem_high = sfpage.snd_base + L4_PAGESIZE;

      if (memmap_owner_page(sfpage.snd_base) == O_FREE)
	{
	  printf("RMGR: WARNING: we've been given page 0x%08x twice!\n",
		 sfpage.snd_base);
	  continue;
	}

      free_size += L4_PAGESIZE;
      memmap_free_page(sfpage.snd_base, O_RESERVED);
    }

  assert(mem_high > 0x200000);	/* 2MB is the minimum for any useful work */

  /* page in memory we own and we reserved before booting; we need to
     do this explicitly as the L4 kernel won't hand it out voluntarily
     (see L4 rerefence manual, sec. 2.7).
     don't care for overlaps with our own area here because we will
     reserve our memory below */
  for (address = l4_trunc_page(mod_range_start);
       address < mod_range_end;
       address += L4_PAGESIZE)
    {
      /* check if we really can touch the memory */
      if (address >= mem_high)
	{
	  printf("RMGR: Can't touch memory at %08x for boot modules. ", 
	          address);
	  if (address > mem_phys)
	    printf("There is no RAM\n" 
		   "      at this address.\n");
	  else
	    {
	      printf("This error occurs because\n"
		     "      the RAM at this address is owned by the kernel.");
	      if (l4_version == VERSION_FIASCO)
		printf(" Fiasco occupies 20%% of the\n"
		       "      physical memory per default.");
	      printf("\n");
	    }
	  boot_fatal();
	  break;
	}
      
      /* force write page fault */
      /* * (volatile int *) address |= 0;  */
      asm volatile ("orl $0, (%0)" 
		    : /* nothing out */
		    : "r" (address)
		    );
      if (memmap_owner_page(address) != O_FREE)
	{
	  memmap_free_page(address, O_RESERVED);
	  free_size += L4_PAGESIZE;
	}
    }

  /* page in adapter space explicitly.  the adapter space will be
     marked "reserved" and special-cased in the pager() in memmap.c */
  for (address = mem_lower * 1024L;
       address < 0x100000L;
       address +=  L4_PAGESIZE)
    {
      error = l4_ipc_call(my_pager, L4_IPC_SHORT_MSG,
			  address, 0,
			  L4_IPC_MAPMSG(0, L4_WHOLE_ADDRESS_SPACE),
			  &sfpage.snd_base, &sfpage.fpage.fpage,
			  L4_IPC_NEVER, &result);
      /* XXX should i check for errors? */
    }
  
  /* reserve memory for ourselves; page in our own memory en-passant */
  for (address = l4_trunc_page(&_start);
       address <= (vm_offset_t) &_end;
       address += L4_PAGESIZE)
    {
      memmap_alloc_page(address, O_RESERVED);
      reserved_size += L4_PAGESIZE;

      /* force write page fault */
      /* * (volatile int *) address |= 0;  */
      asm volatile ("orl $0, (%0)" 
		    : /* nothing out */
		    : "r" (address)
		    );
    }

  /* Reserve memory for Fiasco symbols.
   * Move it to the upper memory limit to prevent memory fragmentation */
  if (fiasco_symbols)
    {
      vm_size_t size = fiasco_symbols_end - fiasco_symbols;
      int i = 0, pages = size / L4_PAGESIZE;

      if (0 == (address = find_free_chunk(pages, 1)))
	{
	  printf("      No space for symbols found, "
	         "disabling Fiasco symbols\n");
	  fiasco_symbols = 0;
	}
      else
	{
	  memcpy((void*)address, (void*)fiasco_symbols, size);
	  fiasco_symbols     = address;
	  fiasco_symbols_end = address+size;
	  for (i=0; i<pages; i++, address += L4_PAGESIZE)
	    {
	      memmap_alloc_page(address, O_RESERVED);
	      reserved_size += L4_PAGESIZE;
	    }
	  asm ("int $3 ; cmpb $30,%%al" 
	        : : "a"(fiasco_symbols)	/* offset symbols */, 
		    "b"(0) 		/* task number */,
		    "c"(1)		/* 1=symbols, 2=lines */);
	}
    }

  /* Reserve memory for Fiasco symbols.
   * Move it to the upper memory limit to prevent memory fragmentation */
  if (fiasco_lines)
    {
      vm_size_t size = fiasco_lines_end - fiasco_lines;
      int i = 0, pages = size / L4_PAGESIZE;

      if (0 == (address = find_free_chunk(pages, 1)))
	{
	  printf("      No space for lines found, "
	         "disabling Fiasco lines\n");
	  fiasco_lines = 0;
	}
      else
	{
	  memcpy((void*)address, (void*)fiasco_lines, size);
	  fiasco_lines     = address;
	  fiasco_lines_end = address+size;
	  for (i=0; i<pages; i++, address += L4_PAGESIZE)
	    {
	      memmap_alloc_page(address, O_RESERVED);
	      reserved_size += L4_PAGESIZE;
	    }
	  asm ("int $3 ; cmpb $30,%%al" 
	        : : "a"(fiasco_lines)	/* offset symbols */, 
		    "b"(0) 		/* task number */,
		    "c"(2)		/* 1=symbols, 2=lines */);
	}
    }

  if (fiasco_symbols || fiasco_lines)
    printf("RMGR: L4 ");
  if (fiasco_symbols)
    printf("symbols at %08x-%08x (%dkB), ",
	fiasco_symbols, fiasco_symbols_end, 
	(fiasco_symbols_end-fiasco_symbols)/1024);
  if (fiasco_lines)
    printf("lines at %08x-%08x (%dkB)",
	fiasco_lines, fiasco_lines_end,
	(fiasco_lines_end-fiasco_lines)/1024);
  if (fiasco_symbols || fiasco_lines)
    putchar('\n');

  /* reserve address 0x1000 for the kernel info page */
  /* we assume the L4 kernel is loaded to 0x1000, creating a virtual
     memory area which doesn't contain mapped RAM so we can use it as
     a scratch area.  this is at least consistent with the L4
     reference manual, appendix A.  however, we flush the page anyway
     before attempting to map something in there, just in case. */
  l4_info = (l4_kernel_info_t *) 0x1000;
  l4_fpage_unmap(l4_fpage((l4_umword_t) l4_info, L4_LOG2_PAGESIZE, 0, 0),
		 L4_FP_FLUSH_PAGE|L4_FP_ALL_SPACES);
  error = l4_ipc_call(my_pager, L4_IPC_SHORT_MSG,
		      1, 1,
		      L4_IPC_MAPMSG((l4_umword_t) l4_info, L4_LOG2_PAGESIZE),
		      &sfpage.snd_base, &sfpage.fpage.fpage,
		      L4_IPC_NEVER, &result);
  if (error)
    panic("RMGR: can't map kernel info page: IPC error 0x%x", error);
  memmap_alloc_page((l4_umword_t) l4_info, O_RESERVED);
  reserved_size += L4_PAGESIZE;

  assert(l4_info->magic == L4_KERNEL_INFO_MAGIC);

  /* the non-RAM high-memory superpages are all free for allocation */
  for (address = 0x8000; address < 0x10000;
       address += L4_SUPERPAGESIZE/0x10000) /* scaled by 0x10000**-1 to
					       prevent overflow */
    {
      memmap_free_superpage(address * 0x10000, O_RESERVED);
    }


  printf("RMGR: total RAM size = %6d KB (reported by bootloader)\n"
	 "              received %6ld KB RAM from sigma0\n"
	 "                       %6ld KB reserved for RMGR\n", 
	 mb_info.mem_upper + mb_info.mem_lower,
	 free_size/1024L, reserved_size/1024L);
}


static void
init_rmgr_debug_info(void)
{
  if (mb_info.flags & L4UTIL_MB_ELF_SHDR)
    {
      l4_umword_t from_sym=0, to_sym=0;
      l4_umword_t from_lin=0, to_lin=0;

      extract_symbols_from_mbinfo(&mb_info, 4, &from_sym, &to_sym);
      extract_lines_from_mbinfo(&mb_info, 4, &from_lin, &to_lin);
      if (from_sym || from_lin)
	printf("RMGR: My ");
      if (from_sym)
	printf("symbols at %08x-%08x (%dkB), ",
	    from_sym, to_sym, (to_sym-from_sym)/1024);
      if (from_lin)
	printf("lines at %08x-%08x (%dkB)",
	    from_lin, to_lin, (to_lin-from_lin)/1024);
      if (from_sym || from_sym)
	putchar('\n');
    }
}


static void
init_iomap(void)
{
  l4_msgdope_t result;
  l4_fpage_t fp;
  int error;
  unsigned p, ignore;

  /* initialize the IO space to "reserved" */
  iomap_init();

  /* only continue and try to get ports if we are running on FIASCO */
  if(l4_version != VERSION_FIASCO)
    return;

  /* try get the whole IO space at once 
     if this fails try every single port
  */
				/* IO fpage for full IO space */
  fp = l4_iofpage(0, L4_WHOLE_IOADDRESS_SPACE,0);

  error = l4_ipc_call(my_pager, L4_IPC_SHORT_MSG,
		      fp.fpage,
		      0,
		      L4_IPC_IOMAPMSG(0, L4_WHOLE_IOADDRESS_SPACE),
		      &ignore, &fp.fpage,
		      L4_IPC_NEVER, &result);
  /* XXX should i check for errors? */
  
  if(l4_ipc_fpage_received(result) /* got something */
     && fp.iofp.f == 0xf	   /* got IO ports */
     && fp.iofp.iosize == L4_WHOLE_IOADDRESS_SPACE
     && fp.iofp.iopage == 0        /* got whole IO space */
     )
    {
      for(p = 0; p < IO_MAX; p++)
	iomap_free_port(p, O_RESERVED);

      printf("RMGR: received IO ports 0000-ffff\n");
    }
  else
      printf("RMGR: received no IO ports\n");
}


#define alloc_from_page(pageptr, size)				\
({								\
  vm_offset_t ret;						\
  vm_offset_t new_pageptr = ((pageptr) + (size) + 3) & ~3;	\
  if (((pageptr) & L4_PAGEMASK) != (new_pageptr & L4_PAGEMASK))	\
    ret = 0;							\
  else								\
    {								\
      ret = (pageptr);						\
      (pageptr) = new_pageptr;					\
    }								\
  ret;								\
})

static void
start_tasks(void)
{
  unsigned i,i2;
  l4_threadid_t t;
  vm_offset_t address;
  l4util_mb_info_t *mbi;
  const char *name;
  l4util_mb_vbe_mode_t *mbi_vbe_mode;
  l4util_mb_vbe_ctrl_t *mbi_vbe_ctrl;
  exec_info_t exec_info;
  int exec_ret;
  struct exec_task e;
  l4_umword_t from_sym, to_sym, from_lin, to_lin;

  l4_umword_t *sp, entry;

  t = myself;
/*   t.id.nest++; */
  t.id.chief = t.id.task;

  i2 = first_task_module;
  for (i = first_task_module; i < mb_info.mods_count; i++, i2++)
    {
      t.id.task = myself.id.task + 1 + (i - first_task_module);
      check(task_alloc(t.id.task, myself.id.task));

      name = mb_mod_name(i);

      /* register name and id for the RMGR_TASK_GET_ID function */
      mb_mod_ids[i2] = t;

      mod_exec_init(mb_mod, i, mb_info.mods_count);

      e.mod_start = (void *)(mb_mod[i].mod_start);
      e.task_no = t.id.task;
      e.mod_no = i;

      printf("RMGR: ");
      if (l4_version == VERSION_FIASCO)
	printf("\033[36m");
      printf("#%02x: loading \"%s\"\n", t.id.task, name);
      if (l4_version == VERSION_FIASCO)
	printf("\033[m");
      printf("      from %08x-%08x to [ ",
	     mb_mod[i].mod_start, mb_mod[i].mod_end);

      exec_ret = exec_load(alloc_exec_read, alloc_exec_read_exec, 
			   &e, &exec_info);
      printf("]\n");
      
      if (exec_ret)
	{
	  printf("      ERROR: exec_load() failed with code %d\n", exec_ret);
	  boot_error();
	  continue;
	}

      /* pass multiboot info to new task in extra page */
      for (address = 0x3000;	/* skip page 0, 1 and 2 */
	   address < mem_high; 
	   address += L4_PAGESIZE)
	{
	  if (! quota_check_mem(t.id.task, address, L4_PAGESIZE))
            continue;
	  if (memmap_owner_page(address) != O_FREE)
	    continue;
	  if (memmap_alloc_page(address, t.id.task))
	    break;
	}
      assert(address < mem_high);

      /* copy mb_info */
      check((mbi = (l4util_mb_info_t *)
	alloc_from_page(address, sizeof(l4util_mb_info_t))));

      *mbi = mb_info;
      mbi->flags &= L4UTIL_MB_MEMORY | L4UTIL_MB_BOOTDEV | L4UTIL_MB_VIDEO_INFO;
      
      /* copy mb_info->vbe_mode and mb_info->vbe_ctrl */
      if (mb_info.flags & L4UTIL_MB_VIDEO_INFO)
	{
	  vm_offset_t old_address = address;
	  if (mb_info.vbe_mode_info)
	    {
	      if ((mbi_vbe_mode = (l4util_mb_vbe_mode_t*)
		     alloc_from_page(address, 
				     sizeof(l4util_mb_vbe_mode_t))) &&
	          (mbi_vbe_ctrl = (l4util_mb_vbe_ctrl_t*)
		     alloc_from_page(address, 
				     sizeof(l4util_mb_vbe_ctrl_t))))
		{
		  *mbi_vbe_mode = mb_vbe_mode;
		  *mbi_vbe_ctrl = mb_vbe_ctrl;
		  mbi->vbe_mode_info = (vm_offset_t)mbi_vbe_mode;
		  mbi->vbe_ctrl_info = (vm_offset_t)mbi_vbe_ctrl;
		}
	      else
		{
		  printf("      ERROR: can't pass VBE video info "
		         "(needs %d bytes)\n",
			 sizeof(l4util_mb_vbe_ctrl_t)+
			 sizeof(l4util_mb_vbe_mode_t));
		  boot_error();
		  /* remove video information from multiboot info */
		  address = old_address;
		  mbi->flags &= ~L4UTIL_MB_VIDEO_INFO;
		}
	    }
	}
      
      /* copy mb_info->cmdline */
      if (mb_mod[i].cmdline)
	{
	  char *string;
	  vm_size_t len = strlen(name) + 1;

	  if (! (string = (char *) alloc_from_page(address, len)))
	    {
	      printf("      ERROR: command line too long -- truncating\n");
	      boot_error();

	      len = L4_PAGESIZE - 0x120;
	      check(string = (char *) alloc_from_page(address, len));
	    }
	  
	  mbi->flags |= L4UTIL_MB_CMDLINE;
	  mbi->cmdline = (vm_offset_t) string;
	  strncpy(string, name, len);
	  *(string + len - 1) = 0; /* zero-terminate string */
	}

      /* before we start allocating extra mem, first allocate stack space */
      check((sp = (l4_umword_t *) alloc_from_page(address, 0x100)));

      sp = (l4_umword_t *) (0x100 + (vm_offset_t) sp);

      /* if any modules have been passed on to this task, copy module info */
      if (bootquota[t.id.task].mods)
	{
	  l4util_mb_mod_t *m;
	  unsigned mod_i = i + 1;
	  unsigned j = bootquota[t.id.task].mods;

	  i += j;
	  if (i >= mb_info.mods_count)
	    {
	      printf("      ERROR: task configured to use %d modules, "
		     "but there are only %d modules\n"
		     "             left for loading -- not starting %s\n",
		     j, mb_info.mods_count - (i - j) - 1, name);
	      boot_error();
	      break;
	    }

	  if (! (m = (l4util_mb_mod_t *) 
		     alloc_from_page(address, j * sizeof(l4util_mb_mod_t))))
	    {
	      printf("      ERROR: out of boot-info memory space --\n"
		     "             won't start %s",
		     name);
	      boot_error();
	      continue;
	    }
	  
	  mbi->mods_addr = (vm_offset_t) m;
	  mbi->mods_count = j;
	    
	  for (; j--; m++, mod_i++)
	    {
	      vm_offset_t mod_mem;
	      *m = mb_mod[mod_i];

	      if (m->cmdline)
		{
		  vm_size_t len = strlen((char *) (m->cmdline)) + 1;
		  char *string = (char *) alloc_from_page(address, len);

		  if (! string)
		    {
		      printf("      ERROR: not enough room for module string"
			     " --\n"
			     "      not passing string \"%s\"\n", 
			     (char *) (m->cmdline));
		      boot_error();

		      m->cmdline = 0;
		      continue;
		    }

		  strcpy(string, (char *) (m->cmdline));
		  m->cmdline = (vm_offset_t) string;
		}

	      printf("      passing module %s [ 0x%x-0x%x ]\n", 
		     mb_mod_name(mod_i), m->mod_start, m->mod_end);

	      for (mod_mem = l4_trunc_page(m->mod_start);
		   mod_mem < m->mod_end;
		   mod_mem += L4_PAGESIZE)
		{
		  if (! quota_check_mem(t.id.task, address, L4_PAGESIZE))
		    {
		      printf("      ERROR: cannot allocate page at 0x%08x "
			     "for this task (not in quota)\n"
			     "      skipping rest of module\n",
			     mod_mem);
		      m->mod_end = mod_mem;
		      boot_error();
		      break;
		    }

		  if ( ! memmap_alloc_page(mod_mem, t.id.task))
		    {
		      printf("      ERROR: cannot allocate page at 0x%08x "
			     "for this task (owned by 0x%x)\n"
			     "      skipping rest of module\n",
			     mod_mem, memmap_owner_page(mod_mem));
		      m->mod_end = mod_mem;
		      boot_error();
		      break;
		    }
		}

	      i2++;
	      mb_mod_ids[i2].raw = 0;
	    }

	  mbi->flags |= L4UTIL_MB_MODS;
	}

      /* copy task_trampoline PIC code to new task's stack */
      sp -= 1 + ((vm_offset_t) task_trampoline_end
		 - (vm_offset_t) task_trampoline) / sizeof(*sp);
      memcpy(sp, task_trampoline, 
	     (vm_offset_t) task_trampoline_end 
	     - (vm_offset_t) task_trampoline);
      entry = (l4_umword_t) sp;

      /* setup stack for task_trampoline() */
      *--sp = (l4_umword_t) mbi;
      *--sp = exec_info.entry;
      *--sp = 0;		/* fake return address */

      printf("      at entry %08x "
	     "via trampoline page code %08x\n",
	     *(sp + 1), entry);

      from_sym = to_sym = 0;
      if (!no_symbols)
	extract_symbols_from_image((l4_addr_t)e.mod_start, t.id.task, 
				   &from_sym, &to_sym);
      from_lin = to_lin = 0;
      if (!no_lines)
	extract_lines_from_image((l4_addr_t)e.mod_start, t.id.task,
				 &from_lin, &to_lin);
      if (from_sym || from_lin)
	printf("      ");
      if (from_sym)
	printf("symbols at %08x-%08x (%dkB), ",
	    from_sym, to_sym, (to_sym-from_sym)/1024);
      if (from_lin)
	printf("lines at %08x-%08x (%dkB)",
	    from_lin, to_lin, (to_lin-from_lin)/1024);
      if (from_sym || from_lin)
	putchar('\n');
      
      t = l4_task_new(t, bootquota[t.id.task].mcp, 
		      (l4_umword_t) sp, entry, myself);
      
      setup_task(t);
    }

  putchar('\n');
}

/*
 * This function reserves memory occupied by loaded modules in the
 * global memory map
 */
static void reserve_module_memory (vm_offset_t start,
                                   vm_offset_t end,
                                   int taskno)
{
  vm_offset_t address;

  printf("%08x-%08x ", start, end);
  
  for (address = l4_trunc_page(start);
       address < end;
       address += L4_PAGESIZE)  {

    if (! quota_check_mem(taskno, address, L4_PAGESIZE))
      boot_panic("\r\nRMGR: cannot allocate page at 0x%08x: not in quota",
		address);

    if (! memmap_alloc_page(address, taskno))
      boot_panic("\r\nRMGR: cannot allocate page at 0x%08x: owned by 0x%x",
		address, memmap_owner_page(address));
  }
}

/* support for loading tasks from boot modules and allocating memory
   for them */

static int alloc_exec_read_exec(void *handle,
				vm_offset_t file_ofs, vm_size_t file_size,
				vm_offset_t mem_addr, vm_size_t mem_size,
				exec_sectype_t section_type)
{
  struct exec_task *e = (struct exec_task *) handle;

  if (! (section_type & EXEC_SECTYPE_ALLOC))
    return 0;
  
  assert(e->task_no <= O_MAX);

  reserve_module_memory (mem_addr, mem_addr + mem_size, e->task_no);

  return mod_exec_read_exec(handle, file_ofs, file_size, 
			    mem_addr, mem_size, section_type);
}

static int alloc_exec_read(void *handle, vm_offset_t file_ofs,
			   void *buf, vm_size_t size,
			   vm_size_t *out_actual)
{
  return mod_exec_read(handle, file_ofs, buf, size, out_actual);
}

/* module loading support -- also used by startup.c to load the L4
   kernel -- therefore exported from here */

vm_offset_t mod_range_start, mod_range_end;
l4util_mb_mod_t *mod_range_start_mod, *mod_range_end_mod;

void mod_exec_init(l4util_mb_mod_t *mod, 
		   unsigned mod_first, unsigned mod_count)
{
  unsigned i;

  /* If no modules were loaded, don't reserve anything */
  if (mod_first >= mod_count)
    return;

  /* compute range for modules which we must not overwrite during
     booting the startup tasks */
  mod_range_start = mod[mod_first].mod_start;
  mod_range_end   = mod[mod_first].mod_end + 1;
  for (i = mod_first + 1; i < mod_count; i++)
    {
      if (mod[i].mod_start < mod_range_start)
	mod_range_start = mod[i].mod_start;
      if (mod[i].mod_end > mod_range_end)
	mod_range_end = mod[i].mod_end + 1;
    }

  mod_range_start_mod = mod + mod_first;
  mod_range_end_mod   = mod + mod_count;
}

static void
check_inside_myself(void *handle, vm_offset_t begin, vm_offset_t end)
{
  struct exec_task *e = (struct exec_task *) handle;
 
  if (!(begin >= (vm_offset_t)&_etext && begin <= (vm_offset_t)&_end &&
	end   >= (vm_offset_t)&_etext && end   <= (vm_offset_t)&_end))
    {
      char *name;
      char namebuf[40];

      namebuf[sizeof(namebuf)-1] = 0;
      name = strncpy(namebuf, mb_mod_name(e->mod_no), sizeof(namebuf)-1);
      
      boot_panic("\r\n"
	         "RMGR: not inside: %08x - %08x: %s\n"
                 "            RMGR: %08x - %08x",
                 begin, end, name,
      		 (vm_offset_t) &_start, (vm_offset_t) &_end);
    }
}

static inline
int overlaps_range(vm_offset_t tst_begin, vm_offset_t tst_end, 
                   vm_offset_t mem_begin, vm_offset_t mem_end)
{
  return !((tst_begin <= mem_begin && tst_end <= mem_begin)
        || (tst_begin >= mem_end   && tst_end >= mem_end));
}

/* Returns:
   0         if range does not overlap with any module
   ~0        if range overlaps with rmgr
   ~(modrel) if range overlaps with module mod_range_start_mod + <modrel>,
             this is the current handled module plus (modrel), not
             the abolute number in exec_task.mod_no!
*/
static int
overlaps_myself(vm_offset_t begin, vm_offset_t end)
{
  l4util_mb_mod_t *mod;

  /* check if module overlaps ourself (RMGR) */
  if (overlaps_range(begin, end, (vm_offset_t)&_start, (vm_offset_t)&_end))
    return ~0;
  
  /* check if module overlaps one of the other modules */
  for (mod = mod_range_start_mod; mod != mod_range_end_mod; mod++)
    if (overlaps_range(begin, end, mod->mod_start, mod->mod_end))
      return ~(mod - mod_range_start_mod);

  return 0;
}

static void
check_overlaps_myself(void *handle, vm_offset_t begin, vm_offset_t end)
{
  struct exec_task *e = (struct exec_task *) handle;
  int overlaps;

  if ((overlaps=overlaps_myself(begin, end))
      || (begin >= mem_high) || (end >= mem_high))
    {
      int i;
      char *name;
      char namebuf[49];

      namebuf[sizeof(namebuf)-1] = 0;
      name = strncpy(namebuf, mb_mod_name(e->mod_no), sizeof(namebuf)-1);
      
      if (overlaps)
	{
	  printf("\n"
		 "RMGR: The %s\n"
	    	 "        %02d: %08x-%08x: %s\n"
    		 "      %s\n"
		 "        00: %08x-%08x: RMGR\n",
		 ~overlaps == 0 ? "RMGR" : "module",
		 e->mod_no, begin, end, name,
		 overlaps ? "overlaps with module"
			  : "wants to be loaded beyond physical memory",
		 (vm_offset_t) &_start, (vm_offset_t) &_end);
	}

      for (i=first_task_module; i<mb_info.mods_count; i++)
	{
     	  name = strncpy(namebuf, mb_mod_name(i), sizeof(namebuf)-1);
	  printf("    %s %02d: %08x-%08x: %s\n",
		 ~overlaps + e->mod_no == i ? "==>" : "   ",
	    	 i, mb_mod[i].mod_start, mb_mod[i].mod_end, name);
	}

      putchar('\n');
      if (overlaps)
	{
	  printf("   The reason for this error message is that the modules as "
		 "loaded by GRUB\n"
		 "   interfere with their final position as defined by the "
		 "linker. Most often\n"
		 "   you can solve this problem by inserting the command "
		 "\"modaddr 0x02000000\"\n"
		 "   immediately as next line after the \"kernel=\" line of "
		 "your menu.lst. Note\n"
		 "   that you have to use a special version of GRUB which is "
		 "available from\n");
	  panic("   http://os.inf.tu-dresden.de/drops/download.html");
	}
      else
	{
	  panic("   Physical memory at %08x-%08x.",
	         begin, end, mem_high);
	}
      
    }
}

int mod_exec_read(void *handle, vm_offset_t file_ofs,
		  void *buf, vm_size_t size,
		  vm_size_t *out_actual)
{
  /* because this function is used for reading "metadata" of the executable,
   * we have to check that the access lays inside of our data/text segement */
  check_inside_myself(handle, (vm_offset_t) buf, (vm_offset_t) buf + size);
  
  return simple_exec_read(handle, file_ofs, buf, size, out_actual);
}

int mod_exec_read_exec(void *handle,
		       vm_offset_t file_ofs, vm_size_t file_size,
		       vm_offset_t mem_addr, vm_size_t mem_size,
			  exec_sectype_t section_type)
{
  if (! (section_type & EXEC_SECTYPE_ALLOC))
    return 0;

  /* this function is used to read executable code and data to be copied
   * to the loaded program image, so we have to check that we don't overwrite
   *   a) ourself
   *   b) the loaded modules by GRUB */
  check_overlaps_myself(handle, mem_addr, mem_addr + mem_size);

  return simple_exec_read_exec(handle, file_ofs, file_size, mem_addr, mem_size,
			       section_type);
}

static int set_small(l4_threadid_t t, int num)
{
  if (num < SMALL_MAX && small_alloc(num, t.id.task))
    {
      l4_sched_param_t sched;
      l4_threadid_t s;

      s = L4_INVALID_ID;
      l4_thread_schedule(t, L4_INVALID_SCHED_PARAM, &s, &s, &sched);
      sched.sp.small = small_space_size 
	| (num * small_space_size * 2);
      s = L4_INVALID_ID;
      l4_thread_schedule(t, sched, &s, &s, &sched);
    }
  else
    {
      printf("RMGR: ERROR: can't set small space %d for task 0x%x\n",
	     num, t.id.task);
      boot_error();

      return 0;
    }

  return 1;
}

static int
set_prio(l4_threadid_t t, int num)
{
  l4_sched_param_t sched;
  l4_threadid_t s;
  
  s = L4_INVALID_ID;
  l4_thread_schedule(t, L4_INVALID_SCHED_PARAM, &s, &s, &sched);
  sched.sp.prio = num;
  /* don't mess up anything else than the thread's priority */
  sched.sp.state = 0;
  sched.sp.small = 0;
  s = L4_INVALID_ID;
  l4_thread_schedule(t, sched, &s, &s, &sched);
  
  return 1;
}

