#include <stdio.h>
#include <unistd.h>

#include <l4/sys/syscalls.h>
#include <l4/sys/ipc.h>
#include <l4/sys/kdebug.h>

#include "config.h"
#include "exec.h"
#include "globals.h"
#include "memmap.h"
#include "memmap_lock.h"

/* the interface we define */
#include <l4/rmgr/proto.h>
#include "rmgr.h"

int boot_mem_dump;

#define MGR_BUFSIZE 1024

#define MGR_STACKSIZE 2048
static char mgr_stack[MGR_STACKSIZE];
static void mgr(void) L4_NORETURN;

/*
 *  we better make this static because of limited stack size
 */
static char str_buf[MGR_BUFSIZE];

const char *
owner_name(owner_t owner)
{
  int i = owner-myself.id.task-1+first_task_module;
  char *s, *f;

  if (i<0 || i>=MODS_MAX)
    return "[unknown]";
  
  s = mb_mod_names[i];
  if (!(f = strrchr(s, '/')) || f[1] == '\0')
    f = s;
  else
    f++;

  return f;
}

/* started from init() in startup.c */
void
rmgr(void)
{
  l4_threadid_t pa, pr;
  l4_umword_t dummy;
  l4_msgdope_t result;

  /* we (the main thread of this task) will serve as a pager for our
     children. */

  /* before that, we start threads for other services, e.g. resource
     management */
  pa = my_pager; pr = my_preempter;
  l4_thread_ex_regs(rmgr_super_id,
		    (l4_umword_t) mgr, (l4_umword_t)(mgr_stack + MGR_STACKSIZE),
		    &pr, &pa, &dummy, &dummy, &dummy);

  if (boot_mem_dump)
    {
      l4_ipc_call(rmgr_super_id, 
		  L4_IPC_SHORT_MSG, RMGR_MEM_MSG(RMGR_MEM_INFO), 0,
		  L4_IPC_SHORT_MSG, &dummy, &dummy,
		  L4_IPC_NEVER, &result);
#ifndef FIASCO_UX
      {
        int c;
        printf("\rRMGR: Return continues, Esc panics");
        c = getchar();
        printf("\r%40s\r","");
        if (c == 27)
          panic("RMGR: memory dumped");
      }
#endif	
    }

  /* now start serving the subtasks */
  pager();
}

/* resource manager */

/* defined here because of the lack of a better place */
quota_t quota[TASK_MAX];
owner_t __task[TASK_MAX];
owner_t __small[SMALL_MAX];

l4util_mb_info_t mb_info;
#ifdef FIASCO_UX
l4util_mb_info_t *mb_ptr;
#endif
char mb_cmdline[0x1000];

#ifdef FIASCO_UX
unsigned first_task_module = 2;
#else
unsigned first_task_module = 1;
#endif
unsigned first_not_bmod_task;
unsigned first_not_bmod_free_task;
unsigned first_not_bmod_free_modnr;
unsigned fiasco_symbols;
unsigned fiasco_symbols_end;
unsigned fiasco_lines;
unsigned fiasco_lines_end;
__mb_mod_name_str mb_mod_names[MODS_MAX];
l4_threadid_t mb_mod_ids[MODS_MAX] = { L4_INVALID_ID, };


static void 
mgr(void)
{
  int task;
  l4_proto_t p;
  l4_umword_t arg, d1, d2, mcp;
  void *desc;
  int err, unhandled, do_log;
  l4_threadid_t sender, partner, n, s;
  l4_msgdope_t result;
  l4_sched_param_t sched;

  struct 
    {
      l4_fpage_t fp;
      l4_msgdope_t size_dope;
      l4_msgdope_t send_dope;
      l4_umword_t      dw2[2];
      l4_strdope_t data;
    } msg;

  for (;;)
    {
      msg.size_dope     = L4_IPC_DOPE(2, 1);
      msg.send_dope     = L4_IPC_DOPE(2, 1);
      msg.data.rcv_size = MGR_BUFSIZE;
      msg.data.rcv_str  = (l4_umword_t)str_buf;
      err = l4_ipc_wait(&sender, &msg, &p.request, &arg, L4_IPC_NEVER,
			&result);
      
      while (!err)
	{
	  /* we received a request here */

	  do_log = ((debug_log_mask & (1L << p.proto.proto))
		    && (! (debug_log_types & 2) 
			|| ((debug_log_types >> 16) == p.proto.action)));

	  if (do_log && (debug_log_types & 1))
	    printf("RMGR: Log: rcv from %x:%x: [%x.%02x], r = 0x%x\n",
		   sender.id.task, sender.id.lthread, p.request, arg,
		   result.msgdope);
	  
	  /* who's the allocation unit? */
	  if (L4_IPC_MSG_DECEITED(result))
	    check(l4_nchief(sender, &partner) == L4_NC_INNER_CLAN);
	  else
	    partner = sender;

#if 0 /* currently, this assertion doesn't hold when using RMGR_TASK_CREATE */
	  assert(task_owner(partner.id.task) == myself.id.task);
#endif
	  /* handle the supervisor protocol */
	  
	  desc = L4_IPC_SHORT_MSG;
	  d1 = d2 = (l4_umword_t) -1;

	  unhandled = 0;

          if (partner.id.task > O_MAX)
            {
              /* OOPS.. can't help this sender. */
	      unhandled = 1;
	      goto reply;
            }

	  switch (p.proto.proto)
	    {
	    case RMGR_RMGR:
	      switch (p.proto.action)
		{
		case RMGR_RMGR_PING:
		  /* answer to ping request */
		  d1 = 0; d2 = ~(p.proto.param);
		  break;
		default:
		  unhandled = 1;
		}
	      break;
	    case RMGR_TASK:
	      switch (p.proto.action)
		{
		case RMGR_TASK_DELETE:
		  /* delete a task we've created on behalf of this user */

		case RMGR_TASK_GET:
		  /* transfer task create right for task no
		     p.proto.param to sender */
		  if (p.proto.param >= TASK_MAX)
		    break;
		  
		  if (p.proto.action == RMGR_TASK_GET)
		    {
		      if (!task_alloc(p.proto.param, partner.id.task))
			break;
		    }
		  else
		    {
		      if (task_owner(p.proto.param) != partner.id.task)
			break;
		    }

		  last_pf[p.proto.param].pfa = -1;
		  last_pf[p.proto.param].eip = -1;
			  
		  n = sender;
		  n.id.task = p.proto.param;
		  n.id.chief = myself.id.task; /* XXX I hope the args
						  suffice for l4_task_new() */
		  n = l4_task_new(n, 
				  p.proto.action == RMGR_TASK_GET
				  ? sender.lh.low : myself.lh.low, 
				  0, 0, L4_NIL_ID);

		  if (l4_is_nil_id(n) || p.proto.action == RMGR_TASK_DELETE)
		    {
		      task_free(p.proto.param, partner.id.task);

		      if (l4_is_nil_id(n) && p.proto.action == RMGR_TASK_GET)
			break;	/* failed */
		    }
		  
		  d1 = 0;

		  break;

		case RMGR_TASK_SET_SMALL:
		case RMGR_TASK_SET_PRIO:
		  n.lh.low = arg;

 		  /* try to guess the high word of the thread id */
		  n.lh.high = partner.lh.high;
		  if (! l4_task_equal(n, partner))
		    {
		      /* XXX only works for childs of partner */
		      n.id.chief = partner.id.task;
		      n.id.nest++;
		    }

		  s = L4_INVALID_ID;
		  l4_thread_schedule(n, L4_INVALID_SCHED_PARAM, &s, &s, &sched);
		  if (l4_is_invalid_sched_param(sched)) /* error? */
		    {
		      /* XXX another try... we talk about a child of sender */
		      if (l4_task_equal(sender, partner))
			break;

		      n.id.chief = sender.id.task;
		      n.id.nest = sender.id.nest + 1;

		      s = L4_INVALID_ID;
		      l4_thread_schedule(n, L4_INVALID_SCHED_PARAM, 
					 &s, &s, &sched);

		      if (l4_is_invalid_sched_param(sched)) /* error? */
			break;	/* give up. */
		    }

		  if (p.proto.action == RMGR_TASK_SET_SMALL)
		    {
		      if (! small_space_size || p.proto.param == 0
			  || p.proto.param >= 128/small_space_size
			  || ! small_alloc(p.proto.param, partner.id.task))
			break;

		      sched.sp.small = small_space_size 
			| (p.proto.param * small_space_size * 2);
		      s = L4_INVALID_ID;
		      l4_thread_schedule(n, sched, &s, &s, &sched);
		    }
		  else		/* p.proto.action == RMGR_TASK_SET_PRIO */
		    {
		      if (quota[partner.id.task].log_mcp < p.proto.param)
			break;

		      sched.sp.prio = p.proto.param;
		      s = L4_INVALID_ID;
		      l4_thread_schedule(n, sched, &s, &s, &sched);
		    }

		  d1 = 0;
		  break;

		case RMGR_TASK_CREATE:
		  /* create a task; arg = task number, p.proto.param = mcp,
		     *(l4_umword_t*)str_buf = esp, 
		     *(l4_umword_t*)(str_buf + 4) = eip,
		     *(l4_threadid_t*)(str_buf + 8) = pager */

		  if (result.md.strings != 1)
		    break;	/* must have 8 bytes indirect string data */

		  n = myself;
		  n.id.chief = myself.id.task;
		  /* we use task id + version number to create the task */
		  n.lh.low = arg; 
		  /* n.id.nest++; */
		  task = n.id.task;

		  if (task >= TASK_MAX 
		      || l4_is_nil_id(* (l4_threadid_t*)(str_buf + 8))
		      || !task_alloc(task, partner.id.task))
		    break;

		  mcp = p.proto.param > quota[partner.id.task].log_mcp 
		    ? quota[partner.id.task].log_mcp : p.proto.param;

		  if (do_log)
		    printf("RMGR: Log: creating task=%x.02%x, esp=%x, "
			   "eip=%x, pager=%x.%02x, mcp=%d\n",
			   n.id.task, n.id.lthread,
			   * (l4_umword_t*)str_buf, * (l4_umword_t*)(str_buf + 4),
			   ((l4_threadid_t*)(str_buf + 8))->id.task,
			   ((l4_threadid_t*)(str_buf + 8))->id.lthread,
			   mcp);
		  
		  n = l4_task_new(n, mcp, * (l4_umword_t*)str_buf, 
				  * (l4_umword_t*)(str_buf + 4),
				  * (l4_threadid_t*)(str_buf + 8));

		  if (l4_is_nil_id(n))
		    {
		      task_free(task, partner.id.task);
		      break;
		    }
		  
		  /* XXX for now, just reset all quotas to their
                     minimal value */
		  quota[task].mem.max = 0;
		  quota[task].himem.max = 0;
		  quota[task].task.max = 0;
		  quota[task].small.max = 0;
		  quota[task].irq.max = -1; //XXX: hack, I NEED a irq
		  quota[task].log_mcp = mcp;

		  d1 = n.lh.low;
		  d2 = n.lh.high;

		  break;

		case RMGR_TASK_CREATE_WITH_PRIO:
		  /* create a task; 
		     arg = task number, 
		     p.proto.param = mcp,
		     *(l4_umword_t*)str_buf = esp, 
		     *(l4_umword_t*)(str_buf + 4) = eip,
		     *(l4_threadid_t*)(str_buf + 8) = pager,
		     *(l4_umword_t*)(str_buf + 16) = sched_param */

		  if (result.md.strings != 1)
		    break;	/* must have 8 bytes indirect string data */

		  n = myself;
		  n.id.chief = myself.id.task;
		  /* we use task id + version number to create the task */
		  n.lh.low = arg; 
		  /* n.id.nest++; */
		  task = n.id.task;

		  /* extract schedule parameters */
		  sched.sched_param = *(l4_umword_t*)(str_buf + 16);
		  sched.sp.small = 0;
		  if (quota[partner.id.task].log_mcp < sched.sp.prio)
		    sched.sp.prio = quota[partner.id.task].log_mcp;
		  if (quota[partner.id.task].log_mcp < p.proto.param)
		    p.proto.param = quota[partner.id.task].log_mcp;
		  
		  if (task >= TASK_MAX 
		      || l4_is_nil_id(* (l4_threadid_t*)(str_buf + 8))
		      || !task_alloc(task, partner.id.task))
		    break;

 		  if (do_log)
		    printf("RMGR: Log: creating task=%x.%02x, esp=%x, "
			   "eip=%x, pager=%x.%02x, sched_param: %lx, mcp: %x\n",
			   n.id.task, n.id.lthread,
			   *(l4_umword_t*)str_buf, *(l4_umword_t*)(str_buf+4),
			   ((l4_threadid_t*)(str_buf+8))->id.task,
			   ((l4_threadid_t*)(str_buf+8))->id.lthread,
			   (unsigned long)sched.sched_param, p.proto.param);

		  n = l4_task_new(n, p.proto.param, * (l4_umword_t*)str_buf, 
				  * (l4_umword_t*)(str_buf + 4),
				  * (l4_threadid_t*)(str_buf + 8));

		  if (l4_is_nil_id(n))
		    {
		      task_free(task, partner.id.task);
		      break;
		    }
		  
		  l4_thread_schedule(n, sched, &s, &s, &sched);

		  /* XXX for now, just reset all quotas to their
                     minimal value */
		  quota[task].mem.max = 0;
		  quota[task].himem.max = 0;
		  quota[task].task.max = 0;
		  quota[task].small.max = 0;
		  quota[task].irq.max = 0;
		  quota[task].log_mcp = p.proto.param;

		  d1 = n.lh.low;
		  d2 = n.lh.high;
		  break;
		  
		case RMGR_TASK_GET_ID:
		  /* default is error */
		  d1 = L4_INVALID_ID.lh.low;
	      	  d2 = L4_INVALID_ID.lh.high;
		  if (result.md.strings == 1)
		    {
		      unsigned int i;
		      char *ptr, *end;

		      for (i = first_task_module; i < mb_info.mods_count; i++)
			{
			  for (ptr = end = mb_mod_names[i];
			       *end && *end != ' '; end++)
			    {
			      if (*end == '/')
				ptr = end + 1;
			    }
			  
			  if (!strncmp (ptr, str_buf, end - ptr))
			    {
			      d1 = mb_mod_ids[i].lh.low;
			      d2 = mb_mod_ids[i].lh.high;
			      break;
			    }
			}
		    }
    		  break;
		  
		case RMGR_TASK_SET_ID:
		  /* default is error */
		  d1 = L4_INVALID_ID.lh.low;
		  d2 = L4_INVALID_ID.lh.high;
		  if (result.md.strings == 1)
		    {
		      int a, b;
		      char *ptr, *end;

		      for (a = first_task_module; a < mb_info.mods_count; a++)
			{
			  for (ptr = end = mb_mod_names[a];
			       *end && *end != ' '; end++)
			    {
			      if (*end == '/')
				ptr = end + 1;
			    }
			  
			  if (!strncmp (ptr, str_buf, end - ptr))
			    {
			      /* found in boot_modules */
			      mb_mod_ids[a].lh.low = sender.lh.low;
		    	      mb_mod_ids[a].lh.high = sender.lh.high;
	    		      mb_mod_ids[a].id.task = p.proto.param;
    			      mb_mod_ids[a].id.lthread = 0;
			      mb_mod_ids[a].id.chief = arg;
			      d1 = d2 = 0;
			      break;
			    }
			}
		      /* search in modules we only have the quotas of */
		      for (b = first_not_bmod_task;
			   a < first_not_bmod_free_modnr; a++, b++)
			{
			  for (ptr = end = mb_mod_names[a];
			       *end && *end != ' '; end++)
			    {
			      if (*end == '/')
				ptr = end + 1;
			    }
			  
			  if (!strncmp(ptr, str_buf, end - ptr))
			    {
			      if (task_owner(p.proto.param)==partner.id.task)
				{
				  memcpy(&quota[p.proto.param], &quota[b],
				         sizeof(quota[0]));
				  d1 = d2 = 0;
				  break;
				}
			    }
			}
		    }
		  break;

		case RMGR_TASK_FREE_ALL:
		  d2 = 0;
		  for (d1 = 0; d1 < TASK_MAX; d1++)
		    {
		      if (task_owner(d1) == arg)
			{
			  last_pf[d1].pfa = -1;
			  last_pf[d1].eip = -1;

			  n = sender; /* XXX */
			  n.id.task  = d1;
			  n.id.chief = myself.id.task;
			  n = l4_task_new(n, myself.lh.low, 0, 0, L4_NIL_ID);

			  if (l4_is_nil_id(n))
			    {
			      d2 = -1;
			      break;
			    }

			  task_free(d1, arg);
			}
		    }
		  d1 = 0;
		  break;

		case RMGR_TASK_FREE:
		  /*
		   * p.proto.param == task no
		   */
		  d1 = -1;

		  if (p.proto.param >= TASK_MAX)
		    break;

		  if (!task_free(p.proto.param, partner.id.task))
		    break;

		  d1 = d2 = 0; /* ok */

		  break;

		default:
		  unhandled = 1;
		}
	      break;

	    case RMGR_IRQ:
	      switch (p.proto.action)
		{
		case RMGR_IRQ_GET:
		  /* transfer right to connect to IRQ to client */
		  if (p.proto.param < IRQ_MAX)
		    {
		      if (irq_alloc(p.proto.param, partner.id.task))
			{
			  /* the irq was free -- detach from it */
			  l4_umword_t e1, e2;
			  n = myself;
			  n.id.lthread = LTHREAD_NO_IRQ(p.proto.param);
			  err = l4_ipc_call(n, L4_IPC_SHORT_MSG, 1, 0,
					    L4_IPC_SHORT_MSG, &e1, &e2,
					    L4_IPC_NEVER, &result);
			  if (e1 == 0)
			    d1 = 0; /* success! */
			  else
			    /* fucked! (couldn't attach to irq number) */
			    irq_free(p.proto.param, partner.id.task);
			}
		    }
		  break;

		case RMGR_IRQ_FREE:
		  /* client transfers irq right back to us */
		  if (p.proto.param < IRQ_MAX
		      && irq_free(p.proto.param, partner.id.task))
		    {
		      l4_umword_t e1, e2;
		      n = myself;
		      n.id.lthread = LTHREAD_NO_IRQ(p.proto.param);
		      err = l4_ipc_call(n, L4_IPC_SHORT_MSG, 0, 0,
					L4_IPC_SHORT_MSG, &e1, &e2,
					L4_IPC_NEVER, &result);
		      check(err == 0);

		      if (e1 == 0)
			d1 = 0;	/* success! */
		    }
		  break;

		case RMGR_IRQ_FREE_ALL:
		  /* free all irqs belonging to a client */
		  for (d1 = 0; d1 < IRQ_MAX; d1++)
		    {
		      if ((irq_owner(d1) == arg) && irq_free(d1, arg))
			{
			  l4_umword_t e1, e2;
			  n = myself;
			  n.id.lthread = LTHREAD_NO_IRQ(d1);
		    	  err = l4_ipc_call(n, L4_IPC_SHORT_MSG, 0, 0,
					    L4_IPC_SHORT_MSG, &e1, &e2,
					    L4_IPC_NEVER, &result);
			  check(err == 0);
			}
		    }
		  d1 = 0; /* success */
		  break;

		default:
		  unhandled = 1;
		}
	      break;

	    case RMGR_MEM:
	      /* we must synchronise the access to memmap functions */
	      enter_memmap_functions(RMGR_LTHREAD_SUPER, rmgr_pager_id);
	      switch (p.proto.action)
		{
		case RMGR_MEM_FREE:
		  /* must contain physical address and owner (== task_no) */
		  /* a bit of paranoia */
		  if (arg >= 0x40000000 && arg < 0xC0000000)
		    {
		      arg += 0x40000000;
		      if((sender.id.task != memmap_owner_page(arg))
			  || !memmap_free_superpage(arg, sender.id.task))
			{
			  d1 = -1; /* failure */
			}
		      else 
			{
			  /* we can't unmap page here because the superpage was
			   * granted so sender has to unmap page itself */
			  d1 = 0;
			}
		    }
		  else
		    {
		      if((sender.id.task != memmap_owner_page(arg))
		          ||  !memmap_free_page(arg,sender.id.task))
			  d1 = -1; /* failure */
		      else {
		        /* unmap fpage */
		        l4_fpage_unmap(l4_fpage(arg, L4_LOG2_PAGESIZE,
						L4_FPAGE_RW, L4_FPAGE_MAP),
				       L4_FP_FLUSH_PAGE|L4_FP_OTHER_SPACES);
		        d1 = 0; /* success */
		      }
		    }
		  break;
		  
		case RMGR_MEM_FREE_FP:
		  printf("RMGR: unable to handle free fpage request.\n");
		  /* I don't know how to resolve the user's fpage 
		     into a physical address 
		     */
		  d1 = -1; /* failure */
		  break;
		  
		case RMGR_MEM_RESERVE:
		  /* reserve memory; arg = size_and_align, 
		     *(l4_umword_t*)str_buf = range_low,
		     *(l4_umword_t*)(str_buf + 4) = range_high */
		  if (result.md.strings != 1)
		    break;	/* must have 8 bytes indirect string data */
		  
		  d1 = reserve_range(arg, partner.id.task,
		                     *(l4_umword_t*)str_buf, 
				     *(l4_umword_t*)(str_buf+4));
		  break;
		  
		case RMGR_MEM_INFO:
		  /* show memory info (using simple printf()) */
	     	  show_mem_info();
		  d1 = 0;
		  break;
		  
		case RMGR_MEM_GET_PAGE0:
		  /* explicit map physical page 0 */
		  d1 = 0;
		  d2 = l4_fpage(0, L4_LOG2_PAGESIZE,
				L4_FPAGE_RW, L4_FPAGE_MAP).fpage;
		  desc = L4_IPC_SHORT_FPAGE;
		  break;
		  
		case RMGR_MEM_FREE_ALL:
		  /* free all memory owned by task <arg> */

		  /* scan all physical memory */
		  for (d1 = 0; d1 < MEM_MAX; d1 += L4_SUPERPAGESIZE)
		    {
		      if (memmap_owner_superpage(arg) == arg)
			{
			  /* superpage belongs to exact the one owner */
			  if (memmap_free_superpage(d1, arg))
			    {
			      l4_fpage_unmap(l4_fpage(d1,
						      L4_LOG2_SUPERPAGESIZE,
						      L4_FPAGE_RW,
						      L4_FPAGE_MAP),
				  	     L4_FP_FLUSH_PAGE|
					     L4_FP_OTHER_SPACES);
			    }
			}
		      if (memmap_owner_superpage(arg) == O_RESERVED)
			{
			  /* superpage belongs to more than one owner */
			  for (d2 = d1; 
			       d2 < d1 + L4_SUPERPAGESIZE; 
			       d2 += L4_PAGESIZE)
			    {
			      if ((memmap_owner_page(d2) == arg)
				  && memmap_free_page(d2, arg))
				{
				  l4_fpage_unmap(l4_fpage(d2, 
							  L4_LOG2_PAGESIZE,
							  L4_FPAGE_RW,
							  L4_FPAGE_MAP),
						 L4_FP_FLUSH_PAGE|
						 L4_FP_OTHER_SPACES);
				}
			    }
			}
		    }
		  
	     	  /* scan adapter space */
		  for (d1 = 0x40000000; d1 < 0xC0000000; d1 += L4_SUPERPAGESIZE)
		    {
		      memmap_free_superpage(d1, arg);
		    }

		  d1 = 0; /* success */
		  break;
		  
		default:
		  unhandled = 1;
		}
	      leave_memmap_functions(RMGR_LTHREAD_SUPER, rmgr_pager_id);
	      break;
	    default:
	      unhandled = 1;
	    }

	reply:
	  if (verbose && unhandled)
	    {
	      printf("RMGR: can't handle request=0x%x, data=0x%x "
		     "from thread=%x\n", p.request, arg, sender.lh.low);
	    }

	  
	  if ((do_log && (debug_log_types & 1))
	      || ((verbose || debug) && d1 == (l4_umword_t)-1))
	    {
	      if (d1 != 0 && !do_log) /* error, and request not yet logged */
		printf("RMGR: Log: rcv from %x.%02x: [0x%x, 0x%x], r = 0x%x\n",
		       sender.id.task, sender.id.lthread, p.request, arg,
		       result.msgdope);
		
	      printf("RMGR: Log: snd to   %x.%02x: [0x%x, 0x%x]\n",
		     sender.id.task, sender.id.lthread, d1, d2);
	    }
	  
	  if ((do_log && (debug_log_types & 4))
	      || (debug && d1 == (l4_umword_t)-1))
	    enter_kdebug("rmgr>");
	  
	  /* send reply and wait for next message */
	  msg.size_dope = L4_IPC_DOPE(2, 1);
	  msg.send_dope = L4_IPC_DOPE(2, 0);
	  msg.data.rcv_size = MGR_BUFSIZE;
	  msg.data.rcv_str  = (l4_umword_t)str_buf;
	  err = l4_ipc_reply_and_wait(sender, desc, d1, d2,
				      &sender, &msg, &p.request, &arg,
				      L4_IPC_TIMEOUT(0,1,0,0,0,0), 
				      /* snd timeout = 0 */
				      &result);
	}
    }
}

