/* startup stuff */

/* it should be possible to throw away the text/data/bss of the object
   file resulting from this source -- so, we don't define here
   anything we could still use at a later time.  instead, globals are
   defined in globals.c */

/* OSKit stuff */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <flux/page.h>

/* L4 stuff */
#include <l4/sys/compiler.h>
#include <l4/sys/kernel.h>
#include <l4/sys/kdebug.h>
#include <l4/util/mb_info.h>

/* local stuff */
#include "exec.h"
#include "globals.h"
#include "rmgr.h"
#include "init.h"
#include "memmap.h"
#include "serial.h"

void startup(l4util_mb_info_t *mbi, unsigned int flag);
int  have_hercules(void);  /* in oskit_support.c */

static exec_read_exec_func_t l4_exec_read_exec;
static vm_offset_t l4_low, l4_high;

static void
check_mb_header(unsigned *start, unsigned *end)
{
  unsigned *p;

  if (end-start < 8192/4)
    panic("RMGR: Image too small -- do you use the right Fiasco image?");

  for (p=start; p<start+(8192/4); p++)
    {
      if (*p == 0x1BADB002
  	  && p[0] + p[1] + p[2] == 0)
    	return;
    }

  panic("RMGR: Boot info not found -- do you use the right Fiasco image?");
}

/*****************************************************************************/
/**
 * \brief  Detect L4 Version
 *
 * \return L4 Version:
 *         - #VERSION_L4_V2    old GMD L4 version
 *         - #VERSION_L4_IBM   IBM LN/X version
 *         - #VERSION_FIASCO   Fiasco 
 *         - #VERSION_L4_KA    L4-KA/Hazelnut
 */
/*****************************************************************************/ 
static int
check_l4_version(char * start, char * end)
{
  char *p;

  no_pentium = 0;

  printf("RMGR: kernel at %08x-%08x\n", (unsigned)start, (unsigned)end);

  for (p = start; p < end; p++)
    {
      if (memcmp(p, "L4/486 \346-Kernel", 15) == 0)
	{
	  int m4_ok;

	  printf("RMGR: detected L4/486\n");
	  no_pentium = 1;

	  /* remove old 4M event */
	  m4_ok = 0;
	  for (p=start; p<end; p++)
	    {
	      if (memcmp(p, "\314\353\0024M", 5) == 0)
		{
		  p[0] = 0x90; /* replace "int $3" by "nop" */
		  m4_ok=1;
		  break;
		}
	    }
	  if (!m4_ok)
	    printf("RMGR: 4M sequence not found in L4/486 -- that's OK.\n");
	  return VERSION_L4_V2;
	}

      if (memcmp(p, "L4/Pentium \346-Kernel", 19) == 0)
	{
	  /* it's Jochen's Pentium version */
	  printf("RMGR: detected L4/Pentium\n");	  
	  return VERSION_L4_V2;
	}

      if (memcmp(p, "Lava Nucleus (Pentium)", 22) == 0)
	{
	  /* it's the IBM version */
	  printf("RMGR: detected IBM LN/Pentium\n");
	  return VERSION_L4_IBM;
	}

      if (memcmp(p, "L4-X Nucleus (x86)", 18) == 0)
	{
	  /* it's the IBM X version */
	  printf("RMGR: detected IBM X/Pentium\n");
#if !defined(L4_API_L4X0) && !defined(L4API_l4x0)
	  /* X.0 API, but not adaption enabled, fail */
	  panic("RMGR: *** L4-X uses 32 Bit TIDs, but not adaption enabled ***");
#endif
	  return VERSION_L4_IBM;
	}

      if (memcmp(p, "DD-L4(v2)/x86 microkernel", 21) == 0
	  || memcmp(p, "DD-L4/x86 microkernel", 21) == 0)
	{
	  /* it's the Dresden version */
	  printf("RMGR: detected new-style DD-L4(v2)/Fiasco\n");
#if defined(L4_API_L4X0) || defined(L4API_l4x0)
	  panic("RMGR: *** Fiasco uses 64 Bit TIDs, but 32 Bit TID adaption enabled ***");
#endif
	  check_mb_header((unsigned*)start, (unsigned*)end);
	  return VERSION_FIASCO;
	}

      if (memcmp(p, "DD-L4(x0)/x86 microkernel", 21) == 0)
	{
	  /* it's the Dresden version */
	  printf("RMGR: detected new-style DD-L4(x0)/Fiasco\n");
#if !defined(L4_API_L4X0) && !defined(L4API_l4x0)
	  panic("RMGR: *** Fiasco(x0) uses 32 Bit TIDs, but 64 Bit TID are used ***");
#endif
	  check_mb_header((unsigned*)start, (unsigned*)end);
	  return VERSION_FIASCO;
	}

      if (memcmp(p, "L4/KA", 5) == 0 ||
	  memcmp(p, "L4Ka/Hazelnut", 13) == 0)
				
      {
	  /* it's the Karlsruhe version */
	  printf("RMGR: detected L4Ka/Hazelnut\n");
#if !defined(L4_API_L4X0) && !defined(L4API_l4x0)
	  /* X.0 API, but not adaption enabled, fail */
	  panic("RMGR: *** Hazelnut uses 32 Bit TIDs, but not adaption enabled ***");
#endif
	  return VERSION_L4_KA;
      }

    }

  printf("RMGR: could not detect version of L4 -- assuming Jochen.\n");
  return VERSION_L4_V2;
}

/*****************************************************************************/
/**
 * \brief Find byte sequence
 */
/*****************************************************************************/ 
static void * 
_scan_kernel(const void * sequence, const int length, const int line)
{
  /* checker */
  unsigned char* find_p;
  void * found_p = 0;
  int found_count = 0;
  for(find_p = (unsigned char *) l4_low;
      find_p < (unsigned char *) l4_high;
      find_p++)
    /* check for some code */
    if (memcmp(find_p, sequence, length) == 0)
      { found_p = (void*) find_p; found_count++; };
  switch (found_count)
    {
    case 0:
      printf("RMGR: sequence in line %d not found\n", line);
      break;
    case 1:
      break;
    default:
      printf("RMGR: sequence in line %d found more than once\n", line);
      found_p = 0;
      break;
    };
  return found_p;
};
#define scan_kernel(sequence) \
_scan_kernel(sequence, sizeof(sequence)-1, __LINE__)

extern vm_offset_t mem_high;
extern vm_offset_t mem_phys;

/*****************************************************************************/
/**
 * \brief  Init old GMD kernel
 */
/*****************************************************************************/ 
static void
init_l4_gmd(l4_kernel_info_t * l4i, exec_info_t * exec_info) 
{
  unsigned i;

  /* assertion: the entry point is at the beginning of the kernel info
     page */
  l4i = (l4_kernel_info_t *) exec_info->entry;
  assert(0x00001000 <= l4i->sigma0_eip && l4i->sigma0_eip < 0x000a0000);
  /* XXX can't think of more sanity checks that l4i really points to
     the L4 kernel info page :-( */
  
  /* XXX undocumented */
  l4i->reserved1.low = l4i->reserved1.high = 0;
  for (i = 2; i < 4; i++)
    l4i->dedicated[i].low = l4i->dedicated[i].high = 0;
  
  if (mb_mod[0].cmdline) /* do we have a command line for L4? */
    {
      /* more undocumented stuff: process kernel options */
      if (strstr((char *) mb_mod[0].cmdline, "hercules") && have_hercules())
	{
	  unsigned char * kd_init = 
	    (char *)((vm_offset_t)l4i - 0x1000 + l4i->init_default_kdebug);
	  
	  assert(kd_init[0] == 0xb0 && kd_init[1] == 'a');
	  kd_init[1] = 'h';
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "nowait"))
	{
	  unsigned char * kd_debug;
	  int nowait_ok = 0;
	  
	  for(kd_debug = (unsigned char *) l4_low;
	      kd_debug < (unsigned char *) l4_low + (l4_high - l4_low);
	      kd_debug++)
	    {
	      /* check for machine code 
		 "int $3; jmp 0f; .ascii "debug"; 0:" */
	      if (memcmp(kd_debug, "\314\353\005debug", 8) == 0)
		{
		  kd_debug[0] = 0x90; /* replace "int $3" by "nop" */
		  nowait_ok = 1;
		  break;
		}
	    }
	  assert(nowait_ok);
	}
      
      /* all those serial port related options */
      {
	unsigned short  portaddr = 0x02F8;	/* This is COM2, the default */
	unsigned short  divisor  = 1;	        /* 115200 bps */
	unsigned char * ratetable = 0;
	
	if (strstr((char *) mb_mod[0].cmdline, " -comport"))
	  {
	    unsigned port;
	    unsigned char* p_adr;
	    
	    const unsigned char port_adr_hi[] = {0x00, 0x3F, 0x2F, 0x3E, 0x2E};
	    
	    port = strtoul((strstr((char *) mb_mod[0].cmdline, " -comport")
			    + strlen(" -comport")), NULL, 10);
	    
	    if (port)
	      {
		p_adr = scan_kernel("\270\003\207\057\000");
		if (p_adr)
		  {
		    printf("RMGR: Setting COM%d as L4KD serial port\n", port);
		    p_adr[3] = port_adr_hi[port];
		    portaddr = (port_adr_hi[port] << 4) | 0x8;
		  }
	      }
	    else
	      printf("RMGR: INVALID serial port\n");
	  }
	
	if (strstr((char *) mb_mod[0].cmdline, " -comspeed"))
	  {
	    unsigned rate;
	    
	    rate = strtoul((strstr((char *) mb_mod[0].cmdline, " -comspeed")
			    + strlen(" -comspeed")), NULL, 10);
	    
	    if (rate)
	      {
		/* check for rate table
		   dw 192, dw 96, dw 48, dw 24, ...
		   the 7th entry is used by default */
		ratetable = scan_kernel("\300\000\140\000\060\000\030\000");
		if (ratetable)
		  {
		    printf("RMGR: Setting L4KD serial port rate to %d bps\n", rate);
		    ((short*) ratetable)[7] = 115200/rate;
		    divisor = 115200/rate;
		  }
	      }
	    else
	      printf("RMGR: INVALID serial rate\n");
	  }
	
	if (strstr((char *) mb_mod[0].cmdline, " -VT"))
	  {

	    if (!ratetable)
	      ratetable = scan_kernel("\300\000\140\000\060\000\030\000");
	    
	    if (ratetable)
	      {
		printf("RMGR: Enabling serial terminal at %3x, %d bps\n", portaddr, 115200/divisor);
		/* Uaah, the word before the rate table holds the port address for remote_outchar */
		((short*) ratetable)[-1] = portaddr;
		
		/* initialize the serial port */
		asm("mov $0x83,%%al;add $3,%%dx;outb %%al,%%dx;"
		    "sub $3,%%dx;mov %%bx,%%ax;outw %%ax,%%dx;add $3,%%dx;"
		    "mov $0x03,%%al;outb %%al,%%dx;dec %%dx;inb %%dx,%%al;"
		    "add $3,%%dx;inb %%dx,%%al;sub $3,%%dx;inb %%dx,%%al;"
		    "sub $2,%%dx;inb %%dx,%%al;add $2,%%dx;inb %%dx,%%al;"
		    "add $4,%%dx;inb %%dx,%%al;sub $4,%%dx;inb %%dx,%%al;"
		    "dec %%dx;xor %%al,%%al;outb %%al,%%dx;mov $11,%%al;"
		    "add $3,%%dx;outb %%al,%%dx\n"
		    :
		    /* no output */
		    :
		    "d" (portaddr),
		    "b" (divisor)
		    :
		    "eax"
		    );
		
		if (strstr((char *) mb_mod[0].cmdline, " -VT+"))
		  {
		    unsigned char* p;
		    if ((p = scan_kernel("\203\370\377\165\034\350")))
		      {
			p[-6] = 1;
			printf("RMGR: -VT+ mode enabled\n");
		      };
		  }
	      }

	  }
	
	if (strstr((char *) mb_mod[0].cmdline, " -I+"))
	  {
	    
	    /* idea:
	       in init_timer the idt_gate is set up for the timer irq handler.
	       There we install the kdebug_timer_intr_handler.
	       in kdebug_timer... timer_int is called (saved location of timer_int) */
	    
	    unsigned long timer_int, kdebug_timer_intr_handler;
	    void* real_location;
	    void* tmp_location;
	    unsigned long relocation;
	    
	    unsigned short* new;
	    void* p;
	    void* dest;
	    
	    /* find remote_info_port */
	    tmp_location = scan_kernel("\300\000\140\000\060\000\030\000");
	    if (tmp_location)
	      tmp_location -= 2;
	    
	    /* find reference to remote_info_port */
	    real_location = scan_kernel("\146\201\342\377\017");
	    if (real_location)
	      real_location = (void*) ((unsigned long) (((unsigned short*) real_location)[4]));
	    
	    relocation = tmp_location - real_location;
	    
	    /* find the place where the timer_int is set up */
	    p = scan_kernel("\000\000"	/* this is a bit illegal !!!*/
			    "\303"		/* c3		*/
			    "\263\050"	/* b3 28	*/
			    "\267\000"	/* b7 00	*/
			    "\270"		/* b8		*/
			    );
	    if (p)
	      {
		timer_int = ((unsigned long*)p)[2];
		
		/* find kdebug_timer_intr_handler */
		dest = scan_kernel("\152\200\140\036\152\010"
				   "\037\314\074\014\074\033");
		/* no more comments */
		if (dest)
		  {
		    kdebug_timer_intr_handler = (((unsigned long) dest) - relocation);
		    ((unsigned long*)p)[2] = kdebug_timer_intr_handler;
		    dest = scan_kernel("\015\260\055\314"
				       );
		    if (dest)
		      {
			new = (unsigned short*) ((((unsigned short*)dest)[-5]) + relocation);
			*new = timer_int;
			printf("RMGR: I+ enabled\n");
		      };
		  };
	      };
	  };
      }
      
      /*
	This option disables the patch that eliminates the miscalculation
	of string offsets in messages containing indirect strings
	
	The patch changes calculation of the first string's offset in the
	message buffer (in ipc_strings).
	This offset was calculated
	
	addr(msgbuffer) + offset(dword_2) + 4*number_of_dwords
	
	this is changed to
	addr(msgbuffer) + offset(dword_0) + 4*number_of_dwords
	^^^^^^^
      */
      if (strstr((char *) mb_mod[0].cmdline, " -disablestringipcpatch"))
	printf("RMGR: string IPC patch disabled\n");
      else
	{
	  unsigned char * patchme;
	  patchme = scan_kernel("\024\213\126\004\301\352\015\215\164\226\024");
	  if (patchme)
	    {
	      patchme[0] = 0x0c;
	      patchme[10] = 0x0c;
	      printf("RMGR: string IPC patch applied\n");
	    }
	  else
	    printf("RMGR: string IPC patch not applied - signature not found\n");
	};
      
      /* heavy undocumented stuff, we simply patch the kernel to get
	 access to irq0 and to be able to use the high resolution timer */
      if (strstr((char *) mb_mod[0].cmdline, "irq0"))
	{
	  unsigned char *kd_debug;
	  int irq0_ok = 0;
	  
	  for(kd_debug = (unsigned char *) l4_low;
	      kd_debug < (unsigned char *) l4_low + (l4_high - l4_low);
	      kd_debug++)
	    {
	      /* check for byte sequence 
		 0xb8 0x01 0x01 0x00 0x00 0xe8 ?? ?? ?? ?? 0x61 0xc3 */
	      if (memcmp(kd_debug, "\270\001\001\000\000\350", 6) == 0)
		{
		  if ((kd_debug[10] == 0141) && (kd_debug[11] == 0303)) 
		    {
		      kd_debug[1] = '\0'; /* enable irq0 */
		      irq0_ok = 1;
		      break;
		    }
		}
	    }
	  assert(irq0_ok);
	}
      
      /* boot L4 in real-mode (for old versions) */
      if (strstr((char *) mb_mod[0].cmdline, "boothack"))
	{
	  /* find "cmpl $0x2badb002,%eax" */
	  vm_offset_t dest 
	    = (vm_offset_t) scan_kernel("\075\002\260\255\053");
	  
	  if (dest)
	    exec_info->entry = dest;
	}
    }
}

/*****************************************************************************/
/**
 * \brief  Init IBM nucleus
 */
/*****************************************************************************/ 
static void 
init_ibm_nucleus(l4_kernel_info_t * l4i)
{
  unsigned i;
  
  // for consistency with Dresden we always enter the kernel debugger...
  l4i->kdebug_config = 0x100; 
  
  /* assertion: the entry point is at the beginning of the kernel info
     page */
  
  assert(0x00001000 <= l4i->sigma0_eip && l4i->sigma0_eip < 0x000a0000);
  /* XXX can't think of more sanity checks that l4i really points to
     the L4 kernel info page :-( */
  
  l4i->reserved1.low = l4i->reserved1.high = 0;
  for (i = 2; i < 4; i++)
    l4i->dedicated[i].low = l4i->dedicated[i].high = 0;
  
  if (mb_mod[0].cmdline)	/* do we have a command line for L4? */
    {
      /* more undocumented stuff: process kernel options */
      if (strstr((char *) mb_mod[0].cmdline, "hercules"))
	{
	  printf("IBM's Nucleus does not support hercules as setting\n");
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "nowait"))
	{
	  l4i->kdebug_config &= ~0x100;
	}
      
      /* all those serial port related options */
      {
	if (strstr((char *) mb_mod[0].cmdline, " -comport"))
	  {
	    unsigned port;
	    const unsigned char port_adr_hi[] = {0x00, 0x3F, 0x2F, 0x3E, 0x2E};
	    
	    port = strtoul((strstr((char *) mb_mod[0].cmdline, " -comport")
			    + strlen(" -comport")), NULL, 10);
	    
	    if (port)
	      {
		printf("RMGR: Setting COM%d as L4KD serial port\n", port);
		l4i->kdebug_config  |= ((port_adr_hi[port] << 4 | 0x8) << 20) & 
		  0xfff00000;
	      }
	    else
	      printf("RMGR: INVALID serial port\n");
	  }
	
	if (strstr((char *) mb_mod[0].cmdline, " -comspeed"))
	  {
	    unsigned rate;
	    
	    rate = strtoul((strstr((char *) mb_mod[0].cmdline, " -comspeed")
			    + strlen(" -comspeed")), NULL, 10);
	    
	    if (rate)
	      {
		l4i->kdebug_config |= ((115200/rate) << 16) & 0xf0000;
	      }
	    else
	      printf("RMGR: INVALID serial rate\n");
	  }
	
	if (strstr((char *) mb_mod[0].cmdline, " -VT"))
	  {
	    // set to serial
	    if (strstr((char *) mb_mod[0].cmdline, " -VT+"))
	      {
		printf("RMGR: VT-option does not exist - use comspeed/port instead\n");
	      }
	  }
	
      } // end of comport-related stuff
      
      
      if (strstr((char *) mb_mod[0].cmdline, "-tracepages"))
	{
	  unsigned tracepages;
	  
	  tracepages = strtoul((strstr((char *) mb_mod[0].cmdline, " -tracepages")
				+ strlen(" -tracepages")), NULL, 10);
	  if (tracepages) 
	    {
	      l4i->kdebug_config |= (tracepages & 0xff);
	      printf("RMGR: debugger configured with %d trace pages\n", 
		     tracepages);
	    }
	  else 
	    printf("RMGR: invalid tracepage number\n");
	  
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "-lastdebugtask"))
	{
	  unsigned ldtask;
	  
	  ldtask = strtoul((strstr((char *) mb_mod[0].cmdline, " -lastdebugtask")
			    + strlen(" -lastdebugtask")), NULL, 10);
	  if (ldtask) 
	    {
	      l4i->kdebug_permission |= (ldtask & 0xff);
	      printf("RMGR: debugger can be invoked by task 0 to %d",
		     ldtask);
	    }
	  else 
	    printf("RMGR: invalid debug task number\n");
	  
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "-d-no-mapping"))
	{
	  l4i->kdebug_config &= ~0x100;
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "-d-no-registers"))
	{
	  l4i->kdebug_config &= ~0x200;
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "-d-no-memory"))
	{
	  l4i->kdebug_config &= ~0x400;
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "-d-no-modif"))
	{
	  l4i->kdebug_config &= ~0x800;
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "-d-no-io"))
	{
	  l4i->kdebug_config &= ~0x1000;
	}
      
      if (strstr((char *) mb_mod[0].cmdline, "-d-no-protocol"))
	{
	  l4i->kdebug_config &= ~0x1000;
	}	
    }
  else 
    printf("no configuration string for the nucleus\n");
  
  printf("RMGR: kernel-debug config: %x, %x\n", 
	 l4i->kdebug_config, l4i->kdebug_permission);
}

/*****************************************************************************/
/**
 * \brief  Init Fiasco
 */
/*****************************************************************************/ 
static void
init_fiasco(l4_kernel_info_t ** l4i)
{
  static char cmdline[256];
  int l;
  
  /* use an info page prototype allocated from our address space;
     this will be copied later into the kernel's address space */
  static char proto[L4_PAGESIZE];

  memset(proto, 0, L4_PAGESIZE);
  *l4i = (l4_kernel_info_t *) proto;
  
  /* append address of kernel info prototype to kernel's command
     line */
  if (mb_mod[0].cmdline)
    {
      l = strlen((char *) mb_mod[0].cmdline);
      /* make sure it fits into cmdline[] */
      assert(l < sizeof(cmdline) - 40);
      strcpy(cmdline, (char *) mb_mod[0].cmdline);
    }
  else
    l = 0;
  
  sprintf(cmdline + l, " proto=0x%x", (vm_offset_t) proto);
  mb_mod[0].cmdline = (vm_offset_t) cmdline;
}

/*****************************************************************************/
/**
 * \brief  Init L4Ka/Hazelnut
 */
/*****************************************************************************/ 
static void
init_hazelnut(l4_kernel_info_t ** l4i)
{
  unsigned char * p;

  //no_pentium = 1;
  *l4i = 0;

  printf("RMGR: find kernel info page...\n");
  for (p = (unsigned char *) (l4_low & 0xfffff000); 
       p < (unsigned char *) l4_low + (l4_high - l4_low); 
       p += 0x1000)
    {
      l4_umword_t magic = L4_KERNEL_INFO_MAGIC;
      if (memcmp(p, &magic, 4) == 0)
	{
	  *l4i = (l4_kernel_info_t *) p;
	  printf("RMGR: found kernel info page at %p\n", p);
	  break;
	}
    }
  assert(*l4i);

  (*l4i)->reserved0.low = l4_low;
  (*l4i)->reserved0.high = l4_high;
}

/*****************************************************************************/
/**
 * \brief  Startup, started from crt0.S
 */
/*****************************************************************************/ 
void
startup(l4util_mb_info_t *mbi, unsigned int flag)
{
  unsigned i;
  l4_kernel_info_t * l4i;
  l4util_mb_info_t l4_mbi;
  exec_info_t exec_info;
  struct exec_task e;
  __superpage_t *s;

  vm_offset_t sigma0_low = 0, sigma0_high = 0, sigma0_start = 0, 
    sigma0_stack = 0; /* zero-initialized to avoid spurious warnings */
  int have_sigma0 = 0;

  vm_offset_t roottask_low = 0, roottask_high = 0, roottask_start = 0, 
    roottask_stack = 0; /* zero-initialized to avoid spurious warnings */
  int have_roottask = 0;
  unsigned mem_upper, mem_lower;

  if (flag == L4UTIL_MB_VALID && (mbi->flags & L4UTIL_MB_CMDLINE))
    {
      int comport = 2;
      char *s;

      if ((s = strstr((char *) mbi->cmdline, " -comport=")))
	comport = strtoul(s+10,0,0);
      
      if (strstr((char *) mbi->cmdline, " -serial"))
	com_cons_init(comport);
    }

  /* we're not an L4 task yet -- we're just a GRUB-booted kernel */

  assert(flag == L4UTIL_MB_VALID); /* we need to be multiboot-booted */

  assert(mbi->flags & L4UTIL_MB_MODS); /* we need at least two boot modules: */
  assert(mbi->mods_count >= 2);	/* 1 = L4 kernel, 2 = first user task */

  assert(mbi->mods_count <= MODS_MAX);
  
  /* print RMGR messages on hercules screen? */
  if ((mbi->flags & L4UTIL_MB_CMDLINE)
     && strstr((char *) mbi->cmdline, " -hercules")
     && have_hercules())
    hercules = 1;

  /* Guess the amount of installed memory from the bootloader. We already
   * need that value here because we have to check for a valid L4 kernel.
   * Later (in init.c), mem_high is determined from the sigma0 server. */
  if (mbi->flags & L4UTIL_MB_MEMORY)
    {
      mem_upper = mbi->mem_upper;
      mem_lower = mbi->mem_lower;
    }
  else
    {
      mem_lower = 639;
      mem_upper = 64*1024;
    }

  /* maxmem= parameter? */
  if ((mbi->flags & L4UTIL_MB_CMDLINE)
      && strstr((char *) mbi->cmdline, " -maxmem="))
    {
      mem_upper = 1024 *
	(strtoul((strstr((char *) mbi->cmdline, " -maxmem=") +
		  strlen(" -maxmem=")), NULL, 10) - 1); 
    }

  /* check if more memory than we can handle */
  if (mem_upper > MEM_MAX / 1024 - 1024)
    mem_upper = MEM_MAX / 1024 - 1024;

  if (mbi->mem_upper != mem_upper)
    printf("RMGR: limiting memory to %d MB (%d KB)\n",
	mem_upper/1024+1, mem_upper+1024);

  /* ensure that we have a valid multiboot info structure */
  mbi->mem_upper = mem_upper;
  mbi->mem_lower = mem_lower;
  mbi->flags     |= L4UTIL_MB_CMDLINE;
  
  mem_high = mem_phys = 1024 * (mbi->mem_upper + 1024);
  
  /* we cannot use memset for __memmap if size of owner_t is not 1 */
  assert(sizeof(owner_t) == 1);
  
  /* We initialize the memmap here to observe loading tasks before starting
   * the L4 kernel. Later in init.c, we re-initialize the memmap */
  memset(__memmap, O_FREE, MEM_MAX/L4_PAGESIZE);
  for (s = __memmap4mb, i = 0; i < SUPERPAGE_MAX; i++, s++)
    *s = (__superpage_t) { 0, O_RESERVED, 0 };
  /* reserve BIOS area + graphics memory */
  memset(__memmap + mbi->mem_lower*1024/L4_PAGESIZE, O_RESERVED,
         ((1024*1024 - mbi->mem_lower*1024)/L4_PAGESIZE));
  /* reserve area beyond main memory */
  memset(__memmap + mem_high/L4_PAGESIZE, O_RESERVED,
         ((MEM_MAX - mem_high)/L4_PAGESIZE));
  /* reserve our image */
  memset(__memmap + (unsigned)&_start/L4_PAGESIZE, O_RESERVED,
         ((round_page((unsigned)&_end) - 
	   trunc_page((unsigned)&_start))/L4_PAGESIZE));

  /* copy Multiboot data structures we still need to a safe place
     before playing with memory we don't own and starting L4 */

  /* rmgr.c defines variables holding copies */
  if (mbi->flags & L4UTIL_MB_VIDEO_INFO)
    /* extended VIDEO information available */
    {
      l4util_mb_info_t *gmbi = (l4util_mb_info_t *)mbi;
      
      printf("RMGR: Using extended multiboot info\n");
      
      /* copy extended multiboot info structure */
      memcpy(&mb_info, mbi, sizeof(l4util_mb_info_t));
      
      /* copy VBE mode info structure */
      if (gmbi->vbe_mode_info)
	{
	  memcpy(&mb_vbe_mode, (void*)(gmbi->vbe_mode_info),
	         sizeof(l4util_mb_vbe_mode_t));
	  mb_info.vbe_mode_info = (vm_offset_t) &mb_vbe_mode;
	}
      
      /* copy VBE controller info structure */
      if (gmbi->vbe_ctrl_info)
	{
	  memcpy(&mb_vbe_ctrl, (void*)(gmbi->vbe_ctrl_info),
	         sizeof(l4util_mb_vbe_ctrl_t));
	  mb_info.vbe_ctrl_info = (vm_offset_t) &mb_vbe_ctrl;
	}
      
      mb_info.flags &= (L4UTIL_MB_MEMORY    | L4UTIL_MB_BOOTDEV  |
	                L4UTIL_MB_CMDLINE   | L4UTIL_MB_MODS     |
			L4UTIL_MB_AOUT_SYMS | L4UTIL_MB_ELF_SHDR |
			L4UTIL_MB_MEM_MAP   | L4UTIL_MB_VIDEO_INFO);
    }
  
  else
    /* extended VIDEO information not available, perhaps we use the old GRUB
     * so only copy the reliable information */
    {
      /* copy simple (old) multiboot structure */
      memcpy(&mb_info, mbi, 
	     (l4_umword_t)&mbi->vbe_ctrl_info -(l4_umword_t)&(mbi->flags));
      mb_info.flags &= (L4UTIL_MB_MEMORY    | L4UTIL_MB_BOOTDEV  |
			L4UTIL_MB_CMDLINE   | L4UTIL_MB_MODS     |
			L4UTIL_MB_AOUT_SYMS | L4UTIL_MB_ELF_SHDR |
			L4UTIL_MB_MEM_MAP);
    }
  
  /* copy module descriptions */
  memcpy(mb_mod, (void *)mbi->mods_addr, 
	 mbi->mods_count * sizeof (l4util_mb_mod_t));
  mb_info.mods_addr = (vm_offset_t) mb_mod;
  
  for (i = 0; i < mbi->mods_count; i++)
    {
      char * name = (char *) mb_mod[i].cmdline;

      strncpy(mb_mod_names[i], name ? name : "[unknown]", MOD_NAME_MAX);
      mb_mod_names[i][MOD_NAME_MAX-1] = '\0'; /* terminate string */
      if (name)
	mb_mod[i].cmdline = (vm_offset_t) mb_mod_names[i];
    }

  if (mb_info.flags & L4UTIL_MB_CMDLINE)
    {
      strncpy(mb_cmdline, (char *) (mb_info.cmdline), sizeof(mb_cmdline));
      mb_cmdline[sizeof(mb_cmdline) - 1] = 0;
      mb_info.cmdline = (vm_offset_t) mb_cmdline;
    }				    
  /* shouldn't touch original Multiboot parameters after here */

  /* compute range for modules which we must not overwrite during
     booting the startup tasks */
  mod_exec_init(mb_mod, 0, mb_info.mods_count);

  /* lets look if we have a user-specified sigma0 */
  if ((mb_info.flags & L4UTIL_MB_CMDLINE)
      && strstr((char *) mb_info.cmdline, " -sigma0"))
    {
      /* we define a small stack for sigma0 here -- it is used by L4
         for parameter passing to sigma0.  however, sigma0 must switch
         to its own private stack as soon as it has initialized itself
         because this memory region is later recycled in init.c */
      static char sigma0_init_stack[64]; /* XXX hardcoded */

      l4_low = 0xffffffff;
      l4_high = 0;

      /* need one more module -- sigma0 */
      assert(mbi->mods_count >= (first_task_module + 2));

      printf("RMGR: loading %s\n", 
	     mb_mod[1].cmdline 
	       ? (char *) mb_mod[1].cmdline 
	       : "[SIGMA0]");

      e.mod_start = (void *)(mb_mod[first_task_module].mod_start);
      e.task_no = first_task_module;
      e.mod_no = first_task_module;
      exec_load(mod_exec_read, l4_exec_read_exec, &e, &exec_info);

      sigma0_low   = l4_trunc_page(l4_low);
      sigma0_high  = l4_round_page(l4_high);
      sigma0_start = exec_info.entry;
      sigma0_stack = 
	(vm_offset_t)(sigma0_init_stack + sizeof(sigma0_init_stack));

      have_sigma0 = 1;
      first_task_module++;
    }

    /* lets look if we want to use rmgr just for loading and configuring the kernel */
  if ((mb_info.flags & L4UTIL_MB_CMDLINE)
      && strstr((char *) mb_info.cmdline, "-roottask"))
    {
      /* we define a small stack for roottask here -- it is used by L4
	 for parameter passing to roottask.  however, roottask must switch
	 to its own private stack as soon as it has initialized itself
	 because this memory region is later recycled in init.c */
      static char roottask_init_stack[64]; /* XXX hardcoded */
      
      l4_low = 0xffffffff;
      l4_high = 0;
      
      //assert(mb_info.mods_count >= (first_task_module + 2)); /* need one more module -- roottask */
      
      printf("RMGR: loading %s\n", 
	     mb_mod[first_task_module].cmdline 
	       ? (char *) mb_mod[first_task_module].cmdline 
	       : "[ROOTTASK]");

      e.mod_start = (void *)(mb_mod[first_task_module].mod_start);
      e.task_no = first_task_module;
      e.mod_no = first_task_module;
      exec_load(mod_exec_read, l4_exec_read_exec, &e, &exec_info);
      
      roottask_low   = l4_trunc_page(l4_low);
      roottask_high  = l4_round_page(l4_high);
      roottask_start = exec_info.entry;
      roottask_stack = 
	(vm_offset_t)(roottask_init_stack + sizeof(roottask_init_stack));
      
      have_roottask = 1;
      first_task_module++;
    }

  if ((mb_info.flags & L4UTIL_MB_CMDLINE)
      && strstr((char *) mb_info.cmdline, " -symbols"))
    {
      printf("RMGR: loading %s\n", 
	     mb_mod[first_task_module].cmdline 
	        ? (char *) mb_mod[first_task_module].cmdline : "[SYMBOLS]");

      e.mod_start = (void *)mb_mod[first_task_module].mod_start;
      e.task_no = 0;
      e.mod_no = first_task_module;
      exec_load(mod_exec_read, l4_exec_read_exec, &e, &exec_info);

      fiasco_symbols     = mb_mod[first_task_module].mod_start;
      fiasco_symbols_end = l4_round_page(mb_mod[first_task_module].mod_end);
      first_task_module++;
    }

  if ((mb_info.flags & L4UTIL_MB_CMDLINE)
      && strstr((char *) mb_info.cmdline, " -lines"))
    {
      printf("RMGR: loading %s\n", 
	     mb_mod[first_task_module].cmdline 
	        ? (char *) mb_mod[first_task_module].cmdline : "[LINES]");

      e.mod_start = (void *)mb_mod[first_task_module].mod_start;
      e.task_no = 0;
      e.mod_no = first_task_module;
      exec_load(mod_exec_read, l4_exec_read_exec, &e, &exec_info);

      fiasco_lines     = mb_mod[first_task_module].mod_start;
      fiasco_lines_end = l4_round_page(mb_mod[first_task_module].mod_end);
      first_task_module++;
    }

  l4_low = 0xffffffff;
  l4_high = 0;

  /* the first module is the L4 microkernel; load it */
  printf("RMGR: loading kernel %s\n",
	 mb_mod[0].cmdline ? (char *) mb_mod[0].cmdline : "");

  e.mod_start = (void *)mb_mod[0].mod_start;
  e.task_no = 0;
  e.mod_no = 0;
  exec_load(mod_exec_read, l4_exec_read_exec, &e, &exec_info);
  
  /* XXX we don't look at the Multiboot header of the L4 kernel image */

  /* check for L4 version */
  l4_version = check_l4_version((unsigned char *) l4_low, 
				(unsigned char *) l4_low + (l4_high - l4_low));

  if ((l4_version != VERSION_FIASCO) && fiasco_symbols)
    {
      printf("RMGR: L4 version is not Fiasco -- disabling Fiasco symbols\n");
      fiasco_symbols = 0;
    }
  if ((l4_version != VERSION_FIASCO) && fiasco_lines)
    {
      printf("RMGR: L4 version is not Fiasco -- disabling Fiasco lines\n");
      fiasco_lines = 0;
    }

  switch (l4_version)
    {
    case VERSION_L4_V2:
      /* init old GMD-Version */
      l4i = (l4_kernel_info_t *) exec_info.entry;
      init_l4_gmd(l4i, &exec_info);
      break;

    case VERSION_L4_IBM:
      l4i = (l4_kernel_info_t *) exec_info.entry;
      init_ibm_nucleus(l4i); 
      break;
      
    case VERSION_FIASCO:
      init_fiasco(&l4i);
      break;

    case VERSION_L4_KA:
      init_hazelnut(&l4i);
      break;

    default:
      l4i = 0;
      printf("RMGR: could not identify version of the kernel\n");
    }

  /* setup the L4 kernel info page before booting the L4 microkernel:
     patch ourselves into the booter task addresses */
  l4i->root_esp = (l4_umword_t) &_stack;
  l4i->root_eip = (l4_umword_t) init;
  if (l4_version != VERSION_FIASCO)
    {
      l4i->root_memory.low  = l4_trunc_page(&_start);
      l4i->root_memory.high = l4_round_page(&_end);
    }
  else
    {
      l4i->root_memory.low  = l4i->dedicated[0].low  = (l4_umword_t) &_start;
      l4i->root_memory.high = l4i->dedicated[0].high = (l4_umword_t) &_end;
    }

  /* set up sigma0 info */
  if (have_sigma0)
    {
      printf("RMGR: configuring sigma0: %08x-%08x, start: %08x\n",
	     sigma0_low, sigma0_high, sigma0_start);

      l4i->sigma0_eip = sigma0_start;
      l4i->sigma0_memory.low = sigma0_low;
      l4i->sigma0_memory.high = sigma0_high;

      /* XXX UGLY HACK: Jochen's kernel can't pass args on a sigma0
         stack not within the L4 kernel itself -- therefore we use the
         kernel info page itself as the stack for sigma0.  the field
         we use is the task descriptor for the unused "ktest2" task */
      switch (l4_version)
	{
	case VERSION_L4_V2:
	case VERSION_L4_IBM:
	  l4i->sigma0_esp = 0x1060;
	  break;
	case VERSION_FIASCO:
	  l4i->sigma0_esp = sigma0_stack;
	  break;
	}  
    }

  /* set up roottask info */
  if (have_roottask)
    {
      printf("RMGR: configuring roottask: %08x-%08x, start: %08x\n", 
	     roottask_low, roottask_high, roottask_start);
      
      l4i->root_eip = roottask_start;
      l4i->root_memory.low = roottask_low;
      l4i->root_memory.high = roottask_high;
      printf("RMGR: roottask config: 0x%08x,0x%08x,0x%08x,0x%08x\n", 
	     l4i->root_eip, l4i->root_esp,
	     l4i->root_memory.low, l4i->root_memory.high);
    }

  /* reserve memory for the modules */
  mod_exec_init(mb_mod, 1, mb_info.mods_count);

  printf("RMGR: reserve modules memory range: %08x-%08x\n",
	 mod_range_start, mod_range_end);

  l4i->dedicated[1].low  = mod_range_start;
  l4i->dedicated[1].high = mod_range_end;

  if (l4_version != VERSION_FIASCO)
    {
      l4i->main_memory.low = 0;
      l4i->main_memory.high = (mb_info.mem_upper + 1024) * 1024;
      l4i->dedicated[0].low = mb_info.mem_lower * 1024;
      l4i->dedicated[0].high = 1024*1024;
    }

  /* now boot the L4 microkernel in a multiboot-compliant way */
  
  l4_mbi = mb_info;
  assert(l4_mbi.flags & L4UTIL_MB_MEMORY);
  l4_mbi.flags = L4UTIL_MB_MEMORY;
  if (mb_mod[0].cmdline)
    {
      l4_mbi.cmdline = mb_mod[0].cmdline;
      l4_mbi.flags |= L4UTIL_MB_CMDLINE;
    }
  
  printf("RMGR: starting kernel %s @ 0x%08x\n",
	 mb_mod[0].cmdline ? (char *) mb_mod[0].cmdline : "[L4]", 
	 exec_info.entry);

  kernel_running = 1;

  asm volatile
    ("pushl $exit\n\t"
     "pushl %2  \n\t"
     "ret\n\t"
     :				/* no output */
     : "a" (L4UTIL_MB_VALID), "b" (&l4_mbi), "r" (exec_info.entry));

  /*NORETURN*/

}

static int 
l4_exec_read_exec(void * handle,
		  vm_offset_t file_ofs, vm_size_t file_size,
		  vm_offset_t mem_addr, vm_size_t mem_size,
		  exec_sectype_t section_type)
{
  int address;
  
  if (! (section_type & EXEC_SECTYPE_ALLOC))
    return 0;
  
  if (mem_addr < l4_low)
    l4_low = mem_addr;
  if (mem_addr + mem_size > l4_high)
    l4_high = mem_addr + mem_size;

  for (address = mem_addr; address < mem_addr+mem_size; address+=L4_PAGESIZE)
    {
      if (memmap_owner_page(address) != O_FREE)
	panic("RMGR: cannot load binary because address at %08x not free", 
	    address);
    }

  return mod_exec_read_exec(handle, file_ofs, file_size, 
			    mem_addr, mem_size, section_type);
}

