/* Extract symbols information from ELF binary and tell Fiasco where 
 * to find it */

#include <stdio.h>
#include <l4/sys/types.h>
#include <l4/sys/kdebug.h>
#include "exec.h"
#include "elf.h"
#include "globals.h"
#include "init.h"
#include "memmap.h"
#include "symbols.h"

static void
extract_symbols(unsigned elf_image, unsigned sh_num, unsigned sh_entsize,
		unsigned sh_offs, unsigned task_no,
		l4_addr_t *from_sym, l4_addr_t *to_sym)
{
  int i;
  Elf32_Shdr *sh_sym, *sh_str;
  unsigned sz_str, sz_sym;

  if (l4_version != VERSION_FIASCO)
    return;

  for (i=0; i<sh_num; i++)
    {
      Elf32_Sym *sym_symtab;
      const char *sym_strtab;

      sh_sym = (Elf32_Shdr*)(elf_image + sh_offs + i*sh_entsize);
      if (sh_sym->sh_type == SHT_SYMTAB)
	{
	  unsigned num_symtab;
	  Elf32_Sym *sym;
	  unsigned size, len;
	  unsigned pages;
	  char *str;
	  l4_addr_t addr, syms;

	  sh_str     = (Elf32_Shdr*)(elf_image + sh_offs 
					       + sh_sym->sh_link*sh_entsize);
	  sz_str     = sh_str->sh_size;
	  if (elf_image)
	    sym_strtab = (char*)(elf_image + sh_str->sh_offset);
	  else
	    sym_strtab = (char*)sh_str->sh_addr;
	  sz_sym     = sh_sym->sh_size;
	  if (elf_image)
	    sym_symtab = (Elf32_Sym*)(elf_image + sh_sym->sh_offset);
	  else
	    sym_symtab = (Elf32_Sym*)sh_sym->sh_addr;
	  num_symtab = sz_sym/sizeof(Elf32_Sym);
	  size       = 0;

	  // first determine size of symbol table
	  for (i=0, sym=sym_symtab; i<num_symtab; i++, sym++)
	    {
	      const char *s_name = sym_strtab + sym->st_name;

	      // ignore something special
	      if (  (sym->st_shndx >= SHN_LORESERVE)
		  ||(sym->st_value == 0)
		  ||(*s_name == '\0')
		  ||(!memcmp(s_name, "Letext", 6))
		  ||(!memcmp(s_name, "_stext", 6)))
		continue;

	      len = strlen(s_name);
	      if (len>100)
		len = 100;
	      size += 11+len+1;
	    }

	  size += 1; // terminating 0
	  pages = l4_round_page(size)/L4_PAGESIZE;
	  if (0 == (addr = syms = find_free_chunk(pages, 1)))
	    {
	      printf("      No space for symbols left\n");
	      return;
	    }
	  str  = (char*)addr;

	  // now copy symbols into area
	  for (i=0, sym=sym_symtab; i<num_symtab; i++, sym++)
	    {
	      const char *s_name = sym_strtab + sym->st_name;
	      int len;

	      // ignore something special
	      if (  (sym->st_shndx >= SHN_LORESERVE)
		  ||(sym->st_value == 0)
		  ||(*s_name == '\0')
		  ||(!memcmp(s_name, "Letext", 6))
		  ||(!memcmp(s_name, "_stext", 6)))
		continue;

	      str += sprintf(str, "%08x   ", sym->st_value);
	      len = snprintf(str, 100, "%s", s_name);
	      str += len > 99 ? 99 : len;
	      *str++ = '\n';
	      *str   = '\0'; // terminate symbols!
	    }

	  // mark pages as reserved
	  for (i=0; i<pages; i++, addr+=L4_PAGESIZE)
	    memmap_alloc_page(addr, O_RESERVED);

	  // now publish symbols to fiasco
	  asm ("int $3 ; cmpb $30,%%al" : : "a"(syms), "b" (task_no), "c" (1));

	  *from_sym = syms;
	  *to_sym   = syms+L4_PAGESIZE*pages;

	  return;
	}
    }

  return;
}

void
extract_symbols_from_image(l4_addr_t elf_image, unsigned task_no,
			   l4_addr_t *from_sym, l4_addr_t *to_sym)
{
  Elf32_Ehdr *ehdr = (Elf32_Ehdr*)elf_image;

  extract_symbols(elf_image, ehdr->e_shnum, ehdr->e_shentsize,
		  ehdr->e_shoff, task_no, from_sym, to_sym);
}

void
extract_symbols_from_mbinfo(l4util_mb_info_t *mbi, unsigned task_no,
			    l4_addr_t *from_sym, l4_addr_t *to_sym)
{
  extract_symbols(0, mbi->syms.e.num, mbi->syms.e.size, 
		  mbi->syms.e.addr, task_no, from_sym, to_sym);
}

